# -*- coding: utf-8 -*-
import datetime as dt
from flask import request, render_template

from enerpi.base import log
from enerpiweb import app


@app.template_filter('text_date')
def text_date(str_date):
    """
    jinja2 template filter for date conversion from:
        # of days --> today +- #days
        'today' / 'yesterday' --> proper conversion

    :param str_date: :str: datestr
    :return: :str: datestr

    """
    try:
        delta = dt.timedelta(days=int(str_date))
        return (dt.date.today() + delta).strftime('%Y-%m-%d')
    except ValueError:
        if str_date == 'today':
            return dt.date.today().strftime('%Y-%m-%d')
        elif str_date == 'yesterday':
            return (dt.date.today() - dt.timedelta(days=1)).strftime('%Y-%m-%d')
        return dt.date.today().strftime('Err_%Y-%m-%d')


@app.template_filter('ts_strftime')
def ts_strftime(ts):
    """
    strftime jinja2 template filter

    :param ts: :datetime: datetime object
    :return: :str: datestr

    """
    try:
        if (ts.hour == 0) and (ts.minute == 0):
            return ts.strftime('%d/%m/%y')
        return ts.strftime('%d/%m/%y %H:%M')
    except AttributeError as e:
        log('AttributeError en template_filter:ts_strftime -> {}'.format(e), 'error', False)
        return str(ts)


# @app.errorhandler(InvalidAPIUsage)
# def handle_invalid_usage(error):
#     response = jsonify(error.to_dict())
#     response.status_code = error.status_code
#     return response


@app.errorhandler(404)
def page_not_found(e):
    """
    Error 404
    :param e: error

    """
    log('page_not_found: {}, url={}'.format(e, request.url), 'error')
    return render_template('error.html', error_e=e, code=404), 404


@app.errorhandler(500)
def internal_server_error(e):
    """
    Error 500
    :param e: error

    """
    log('INTERNAL_SERVER_ERROR: {}, request={}'.format(e, request), 'error')
    return render_template('error.html', error_e=e, code=500), 500



# -*- coding: utf-8 -*-
"""
* ︎ENERPI AC CURRENT SENSOR
* ANALOG SENSING with RASP.io ANALOG ZERO (Raspberry PI Hat with MCP3008) *

** ENERPI SCHEMATIC:

  ·············                              Rb:= Burden resistor (into the SCT030-030 probe, ~62 Ω)
  ·  ____________________↘○ ANALOG IN        C:= 10µF capacitor
  ·  │    │   ·                              R1, R2:= 2x 10KΩ resistor
  ·  #    ︎│   ·    _______• VREF (3.3V)
  ·  #   Rb   ·    │                         SCT probe connected to A_in (○) & V_divider (○)
  ·  #    │   ·    R1                          * Ideally, R1 = R2, so V_divider ≅ Vref / 2
  ·  │    │   ·    │
  ·  _____│_______↘○__R2__• GND
  ·············    │      │
    SCT030-030     │__||__│
                       C  │
                          ⏚

** LIGHT SENSOR SCHEMATIC:

   •____𝙍____•___LDR___•                    LDR:= Light Resistor
   │         │         │                     R:= 10KΩ resistor
   ⏚      ANALOG     VREF
   GND  	IN       (3.3V)

"""
import os


PRETTY_NAME = "︎⚡⚡ ︎ENERPI AC CURRENT SENSOR ⚡⚡"
DESCRIPTION = 'AC Current Meter for Raspberry PI with GPIOZERO and MCP3008'
BASE_PATH = os.path.abspath(os.path.dirname(__file__))

__version__ = '0.8.8.6'



# -*- coding: utf-8 -*-
from collections import deque
import datetime as dt
from flask import Response, request, render_template, jsonify
import json
import os
import pandas as pd
from threading import Thread, current_thread
from time import sleep, time

from enerpi.base import SENSORS, log
from enerpi.api import enerpi_receiver_generator, enerpi_data_catalog
from enerpiplot.plotbokeh import get_bokeh_version, html_plot_buffer_bokeh
from enerpiweb import app


BUFFER_MAX_SAMPLES = 1800
STREAM_MAX_TIME = 1800
BOKEH_VERSION = get_bokeh_version()

# GLOBAL VARIABLES FOR ENERPI DATA BROADCAST
last_data = {}
buffer_last_data = deque([], maxlen=BUFFER_MAX_SAMPLES)
thread_receiver = None


def stream_is_alive():
    """
    Check enerpi emitter stream
    """
    global last_data
    last = last_data.copy()
    try:
        is_sender_active = (pd.Timestamp.now(tz=SENSORS.TZ) - last[SENSORS.ts_column]) < pd.Timedelta('1min')
    except KeyError:
        # last_ts, last_power, host_logger = last_data['ts'], last_data['power'], last_data['host']
        is_sender_active = False
        last = {'host': '?', SENSORS.main_column: -1, SENSORS.ts_column: pd.Timestamp.now(tz=SENSORS.TZ)}
    return is_sender_active, last


def _format_event_stream(d_msg, timeout_retry=None, msg_id=None):
    if msg_id is not None:
        return 'id: {}\ndata: {}\n\n'.format(msg_id, json.dumps(d_msg))
    if timeout_retry is not None:
        return 'retry: {}\ndata: {}\n\n'.format(timeout_retry, json.dumps(d_msg))
    return 'data: {}\n\n'.format(json.dumps(d_msg))


def _get_dataframe_buffer_data():
    global buffer_last_data
    if len(buffer_last_data) > 0:
        df = pd.DataFrame(list(buffer_last_data))
        try:
            df = df[SENSORS.columns_sampling].set_index(SENSORS.ts_column)
        except KeyError as e:
            log('KeyError "{}" in _get_dataframe_buffer_data'.format(e), 'error')
            df = df[list(filter(lambda x: x in df, SENSORS.columns_sampling))].set_index(SENSORS.ts_column)
        for s in filter(lambda x: x.name in df, SENSORS):
            if s.is_rms:
                df[s.name] = df[s.name].astype(int)
            else:
                df[s.name] = pd.Series(df[s.name] * 100).round(1)
        for c in filter(lambda x: x in df, [SENSORS.ref_rms, SENSORS.ref_mean]):
            if c in df:
                try:
                    df[c] = df[c].astype(int)
                except ValueError:
                    df[c] = df[c].astype(float).astype(int)
        return df
    return None


def _gen_stream_data_bokeh(start=None, end=None, last_hours=None,
                           rs_data=None, rm_data=None, use_median=False, kwh=False):
    tic = time()
    if start or end or last_hours:
        cat = enerpi_data_catalog(check_integrity=False)
        if kwh:
            df = cat.get_summary(start=start, end=end, last_hours=last_hours)
        else:
            df = cat.get(start=start, end=end, last_hours=last_hours)
            if (df is not None) and not df.empty:
                df = df[SENSORS.columns_sampling]
                # TODO Revisar formatos de analog mean!
                # SENSORS
                # df.ldr = df['ldr'].astype(float) / 10.
                if last_hours is not None:
                    df_last_data = _get_dataframe_buffer_data()
                    if df_last_data is not None:
                        df_last_data = df_last_data.tz_localize(None)
                        df = pd.DataFrame(pd.concat([df, df_last_data])
                                          ).sort_index().drop_duplicates(keep='last')
                df = cat.resample_data(df, rs_data=rs_data, rm_data=rm_data, use_median=use_median)
    else:
        df = _get_dataframe_buffer_data()
    toc_df = time()
    if df is not None and not df.empty:
        try:
            script, divs, version = html_plot_buffer_bokeh(df, is_kwh_plot=kwh)
            toc_p = time()
            log('Bokeh plot gen in {:.3f} s; pd.df in {:.3f} s.'.format(toc_p - toc_df, toc_df - tic), 'debug', False)
            yield _format_event_stream(dict(success=True, b_version=version, script_bokeh=script, bokeh_div=divs[0],
                                            took=round(toc_p - tic, 3), took_df=round(toc_df - tic, 3)))
        except Exception as e:
            msg = 'ERROR en: BOKEH PLOT: {} [{}]'.format(e, e.__class__)
            print(msg)
            yield _format_event_stream(dict(success=False, error=msg))
    else:
        msg = ('No hay datos para BOKEH PLOT: start={}, end={}, last_hours={}, '
               'rs_data={}, rm_data={}, use_median={}, kwh={}<br>--> DATA: {}'
               .format(start, end, last_hours, rs_data, rm_data, use_median, kwh, df))
        # print(msg.replace('<br>', '\n'))
        yield _format_event_stream(dict(success=False, error=msg))
    yield _format_event_stream('CLOSE')


###############################
# BROADCAST RECEIVER
###############################
def _init_receiver_thread():
    log('**INIT_BROADCAST_RECEIVER en PID={}'.format(os.getpid()), 'info', False)
    gen = enerpi_receiver_generator()
    count = 0
    while True:
        try:
            last = next(gen)
            global last_data
            global buffer_last_data
            last_data = last.copy()
            buffer_last_data.append(last_data)
            # print('DEBUG STREAM: last_data --> ', last_data)
        except StopIteration:
            log('StopIteration on counter={}'.format(count), 'debug', False)
            if count > 0:
                sleep(2)
                gen = enerpi_receiver_generator()
            else:
                log('Not receiving broadcast msgs. StopIteration at init!', 'error', False)
                break
        count += 1
        sleep(.5)
    log('**BROADCAST_RECEIVER en PID={}, thread={}. CLOSED on counter={}'
        .format(os.getpid(), current_thread(), count), 'warn', False)


@app.before_first_request
def _init_receiver():
    global thread_receiver
    # Broadcast receiver
    if not thread_receiver:
        thread_receiver = Thread(target=_init_receiver_thread)
        thread_receiver.setDaemon(True)
        thread_receiver.start()


###############################
# STREAM REAL-TIME SENSORS DATA
###############################
@app.route('/api/last')
def last_value_json():
    """
    Return last received msg from ENERPI logger in JSON

    :return: :json: msg
    """
    global last_data
    send = last_data.copy()
    if SENSORS.ts_column in send:
        send[SENSORS.ts_column] = send[SENSORS.ts_column].strftime('%Y-%m-%d %H:%M:%S.%f')[:-4]
    return jsonify(send)


@app.route("/api/stream/realtime", methods=["GET"])
def stream_sensors():
    """
    Stream real-time data as it is received from ENERPI broadcast.
    Used for real-time monitoring of ENERPI values.

    :return: SSE stream response

    """
    def _gen_stream_last_values():
        last_ts = dt.datetime.now(tz=SENSORS.TZ) - dt.timedelta(minutes=1)
        count = 0
        tic = time()
        while time() - tic < STREAM_MAX_TIME:
            global last_data
            if SENSORS.ts_column in last_data and last_data[SENSORS.ts_column] > last_ts:
                send = last_data.copy()
                last_ts = send[SENSORS.ts_column]
                send[SENSORS.ts_column] = send[SENSORS.ts_column].strftime('%Y-%m-%d %H:%M:%S.%f')[:-4]
                yield _format_event_stream(send)
                count += 1
                sleep(.5)
            else:
                sleep(.1)
        yield _format_event_stream('CLOSE')

    return Response(_gen_stream_last_values(), mimetype='text/event-stream')


@app.route('/api/stream/bokeh', methods=['GET'])
@app.route('/api/stream/bokeh/from/<start>', methods=['GET'])
@app.route('/api/stream/bokeh/from/<start>/to/<end>', methods=['GET'])
@app.route('/api/stream/bokeh/last/<last_hours>', methods=['GET'])
def bokeh_buffer(start=None, end=None, last_hours=None):
    """
    Stream the bokeh data to make a plot.
    Used for load/reload plots from user queries.

    :param start: :str: start datetime of data query
    :param end: :str: end datetime of data query
    :param last_hours: :str: query from 'last hours' until now

    :return: SSE stream response

    """
    kwargs = dict(start=start, end=end, last_hours=last_hours)
    if 'kwh' in request.args:
        kwargs.update(kwh=request.args['kwh'] == 'true')
    if 'rs_data' in request.args:
        kwargs.update(rs_data=request.args['rs_data'])
    if 'rm_data' in request.args:
        kwargs.update(rm_data=request.args['rm_data'])
    if 'use_median' in request.args:
        kwargs.update(use_median=request.args['use_median'])
    return Response(_gen_stream_data_bokeh(**kwargs), mimetype='text/event-stream')


###############################
# INDEX WITH REAL TIME MONITOR
###############################
@app.route('/index')
def index():
    """
    Webserver index page with real time monitoring of ENERPI

    """
    global buffer_last_data
    with_last_samples = 'table' in request.args
    return render_template('index.html', include_consumption=True, include_ldr=True,
                           with_last_samples=with_last_samples, last_samples=list(buffer_last_data))



# -*- coding: utf-8 -*-
from flask import request, redirect, url_for, render_template, jsonify
import json
import os

from enerpi.base import log
from enerpi.api import enerpi_data_catalog
from enerpiplot.plotbokeh import get_bokeh_version
from enerpiweb import app, WITH_ML_SUBSYSTEM
from enerpiweb.rt_stream import stream_is_alive


COLOR_BASE = app.config['BASECOLOR']
BOKEH_VERSION = get_bokeh_version()

if WITH_ML_SUBSYSTEM:
    # from enerpiweb.views_labeling import *
    # TODO Integración con las vistas de enerpiprocess (separadas en otro project por dependencias, por ahora)
    pass
# else:
#     @app.route('/learning')
#     def index_learning():
#         return redirect(url_for('control'))


#############################
# INDEX & ROUTES
#############################
# TODO RESTART APP (RELOAD CONFIG!):
@app.route('/reloadserver/')
def reloadserver():
    log('Se procede a intentar reiniciar el servidor', 'debug', True)
    os.system('sudo service uwsgi-emperor restart')
    log('Ha funcionado??', 'debug', True)
    return redirect(url_for('index'))


# TODO Fix layout of control buttons
@app.route('/control')
def control():
    """
    Admin Control Panel with links to LOG viewing/downloading, hdf_stores download, ENERPI config editor, etc.

    """
    is_sender_active, last = stream_is_alive()
    alerta = request.args.get('alerta', '')
    if alerta:
        alerta = json.loads(alerta)
    cat = enerpi_data_catalog(check_integrity=False)
    df = cat.tree
    if df is not None:
        df = df[df.is_cat & df.is_raw].sort_values(by='ts_ini', ascending=False)
        paths_rel = [(os.path.basename(p), t0.strftime('%d/%m/%y'), tf.strftime('%d/%m/%y'), n)
                     for p, t0, tf, n in zip(df['st'], df['ts_ini'], df['ts_fin'], df['n_rows'])]
    else:
        paths_rel = []
    return render_template('control_panel.html',
                           d_catalog={'path_raw_store': os.path.join(cat.base_path, cat.raw_store),
                                      'path_catalog': os.path.join(cat.base_path, cat.catalog_file),
                                      'ts_init': cat.min_ts, 'ts_catalog': cat.index_ts},
                           d_last_msg=last, is_sender_active=is_sender_active, list_stores=paths_rel, alerta=alerta)


@app.route('/api/monitor')
def only_tiles():
    """
    Url route for showing only the real-time monitoring tiles
    (for emmbed in another place, e.g.)

    """
    ldr = request.args.get('ldr', 'true') == 'true'
    consumption = request.args.get('consumption', 'false') != 'false'
    return render_template('only_tiles.html', include_consumption=consumption, include_ldr=ldr)


@app.route('/api/help', methods=['GET'])
def api_help():
    # TODO hacer tabla en html con url rules:
    endpoints = [rule.rule for rule in app.url_map.iter_rules()
                 if rule.endpoint != 'static']
    return jsonify(dict(api_endpoints=endpoints))


@app.route('/api/bokehplot')
def bokehplot():
    """
    Base webpage for query & show bokeh plots of ENERPI data

    """
    return render_template('bokeh_plot.html', url_stream_bokeh=url_for('bokeh_buffer'), b_version=BOKEH_VERSION)


@app.route('/')
def base_index():
    """
    Redirects to 'index', with real-time monitoring of ENERPI values

    """
    return redirect(url_for('index'))


# @app.route('/allroutes')
# def allroutes():
#     routes = []
#     for rule in app.url_map.iter_rules():
#         # To show how to do it, we are going to filter what type of
#         # request are we going to show, for example here we are only
#         # going to use GET requests
#         if "GET" in rule.methods:
#             url = rule.rule
#             routes.append(url)
#     return render_template('sysadmin/routes.html', routes=routes)



# -*- coding: utf-8 -*-
import pandas as pd
from enerpi.hdftscat import HDFTimeSeriesCatalog
from enerpi.base import SENSORS


'''
    @staticmethod
    def process_data(data):
        # Implementar
        return data

    @staticmethod
    def process_data_summary(data):
        # Implementar
        return data, data

    @staticmethod
    def process_data_summary_extra(data):
        # Implementar
        return data, data, None
'''

DELTA_MAX_CALC_CONSUMO_SAMPLE_BFILL = 120  # pd.Timedelta('2min')
# TODO Terminar Doc


def _compress_data(data):
    """
    Down-casting of raw data to minimize disk usage.

    RMS sensor data is reduced to float32
    MEAN sensor data take integer values from 0 to 1000 (int16)
    Ref counters of rms & mean sensors also take integer values (int16)
    :param data: :pd.DataFrame: raw data
    :return: :pd.DataFrame: compressed raw data
    """
    if data is not None:
        if not data.empty:
            data = data.copy().astype('float32')
            data[SENSORS.ref_rms] = data[SENSORS.ref_rms].astype('int16')
            for c in SENSORS.columns_sensors_mean:
                data[c] = pd.Series(1000. * data[c]).round().astype('int16')
    return data


def _process_data(data, append_summary=False):
    """
    Calculate consumption and some stats of enerpi data...

    :param data: :pd.DataFrame: archived data
    :param append_summary: :bool: process summary of data and returns tuple of processed & summary dataframes
    :return:
    """
    consumo = None
    if data is not None and not data.empty and (append_summary or ('high_delta' not in data.columns)):
        data = data.copy()
        cols_rs = [SENSORS.main_column, 'Wh', 'delta', 'high_delta', 'execution']
        data[cols_rs[2]] = pd.Series(data.index).diff().fillna(method='bfill').dt.total_seconds().values
        data[cols_rs[3]] = False
        data[cols_rs[4]] = False
        data.loc[data[cols_rs[2]] > 3, cols_rs[3]] = True
        data.loc[data[cols_rs[2]] > 60, cols_rs[4]] = True
        data.loc[data[cols_rs[2]] > DELTA_MAX_CALC_CONSUMO_SAMPLE_BFILL,
                 cols_rs[2]] = DELTA_MAX_CALC_CONSUMO_SAMPLE_BFILL
        data[cols_rs[1]] = data[SENSORS.main_column] * data.delta / 3600
        if append_summary:
            resampler = data[cols_rs].resample('1h', label='left')
            consumo = pd.DataFrame(resampler[cols_rs[1]].sum().rename('kWh')).fillna(0.).astype('float32')
            consumo /= 1000.
            consumo['t_ref'] = pd.Series(resampler[cols_rs[2]].sum() / 3600).astype('float32')
            consumo['n_jump'] = resampler[cols_rs[3]].sum().fillna(0).astype('int16')
            consumo['n_exec'] = resampler[cols_rs[4]].sum().fillna(0).astype('int32')
            consumo['p_max'] = resampler[SENSORS.main_column].max().round(0).astype('float16')
            consumo['p_mean'] = resampler[SENSORS.main_column].mean().round(0).astype('float16')
            consumo['p_min'] = resampler[SENSORS.main_column].min().round(0).astype('float16')
        data[cols_rs[3]] = data[cols_rs[3]].astype(bool)
        data[cols_rs[4]] = data[cols_rs[4]].astype(bool)
        data.drop([cols_rs[2], cols_rs[1]], axis=1, inplace=True)
        if append_summary:
            return data, consumo
        return data
    elif append_summary:
        return data, None
    return data


class EnerpiCatalog(HDFTimeSeriesCatalog):

    @staticmethod
    def is_raw_data(data):
        if (len(data.columns) == 6) or (data['ldr'].dtype == 'int16'):
            return False
        return True

    @staticmethod
    def process_data(data, is_raw=True):
        if is_raw:
            data = _compress_data(data)
        return _process_data(data, append_summary=False)

    @staticmethod
    def process_data_summary(data):
        return _process_data(data, append_summary=True)

    @staticmethod
    def process_data_summary_extra(data):
        data, consumo = _process_data(data, append_summary=True)
        return data, consumo, None



# -*- coding: utf-8 -*-
# noinspection PyUnresolvedReferences
from enerpi.base import CONFIG, SENSORS, DATA_PATH, FILE_LOGGING, LOGGING_LEVEL, log
# noinspection PyUnresolvedReferences
from enerpi.database import init_catalog, get_ts_last_save, extract_log_file, delete_log_file, HDF_STORE
# noinspection PyUnresolvedReferences
from enerpi.enerpimeter import receiver_msg_generator, enerpi_raw_data, msg_to_dict
# noinspection PyUnresolvedReferences
from enerpi.iobroadcast import get_encryption_key, get_codec


def enerpi_receiver_generator(verbose=False, n_msgs=None):
    """
    Generator of broadcasted values by ENERPI Logger.

    It can be used by any machine in the same network as the ENERPI Logger. It decrypts the encrypted broadcast and
    returns a dict of vars - values.
    Used by the webserver for read & stream real-time values.

    :param verbose: :bool: Log to stdout
    :param n_msgs: :int: # of msgs to receive (unlimited by default).
    :return: :dict:
    """
    gen = receiver_msg_generator(verbose=verbose, n_msgs=n_msgs)
    count = 0
    while True:
        try:
            msg, _t1, _t2 = next(gen)
            yield msg_to_dict(msg)
            count += 1
        except StopIteration:
            log('EXIT from enerpi_receiver_generator. StopIteration in msg #{}'.format(count), 'error', verbose)
            break
    return None


def enerpi_default_config():
    """
    Default configuration for ENERPI Data Catalog, read from INI file.

    :return: :dict: parameters
    """
    conf = {'store': HDF_STORE,
            'DATA_PATH': DATA_PATH,
            'delta': SENSORS.delta_sec_data,
            'window': SENSORS.rms_roll_window_sec,
            'ts': SENSORS.ts_data_ms,
            'LOGGING_LEVEL': LOGGING_LEVEL,
            'FILE_LOGGING': FILE_LOGGING}
    return conf


def enerpi_data_catalog(check_integrity=False, **kwargs):
    """
    Get ENERPI data catalog for access & operation.

    :param check_integrity: :bool: False by default. If true, checks integrity and generates / updates data index.
    :param kwargs: :dict:
    :return: :EnerpiCatalog:
    """
    return init_catalog(check_integrity=check_integrity, **kwargs)


# def get_last_saved_data(path_st=HDF_STORE, get_last_sample=True, verbose=True, n=10):
#     return get_ts_last_save(path_st, get_last_sample=get_last_sample, verbose=verbose, n=n)


# def enerpi_log(log_file=FILE_LOGGING, extract_temps=False, verbose=True):
#     return extract_log_file(log_file, extract_temps=extract_temps, verbose=verbose)



# -*- coding: utf-8 -*-
import locale
# import logging
import numpy as np
import pandas as pd
import bokeh
from bokeh.plotting import figure
from bokeh.embed import components
from bokeh.models import HoverTool, ColumnDataSource, LinearAxis, Range1d  # , NumeralTickFormatter, BoxAnnotation,
from bokeh.io import reset_output


ROUND_W = 500
ROUND_KWH = .5

COLS_DATA = ['power', 'ldr', 'ref']
COLORS_DATA = ['#0CBB43', '#F4D83F', '#8C27D3']
UNITS_DATA = ['W', '%', '']
LABELS_DATA = ['Power', 'LDR', 'Samples']
FMT_TOOLTIP_DATA = ['{0}', '{0.0}', '{0}']

COLS_DATA_KWH = ['kWh', 'p_max', 'p_min', 't_ref']
COLORS_DATA_KWH = ['#8C27D3', '#972625', '#f4af38', '#8C27D3']
UNITS_DATA_KWH = ['kWh', 'W', 'W', '']
LABELS_DATA_KWH = ['Consumption', 'Max Power', 'Min Power', 'Sampled']
FMT_TOOLTIP_DATA_KWH = ['{0.000}', '{0}', '{0}', '{0.000}']

locale.setlocale(locale.LC_ALL, locale.getlocale())
# locale.setlocale(locale.LC_ALL, CONFIG.get('ENERPI_SAMPLER', 'LOCALE', fallback='{}.{}'.format(*locale.getlocale())))

TOOLS = "pan,xwheel_zoom,box_zoom,reset,save,crosshair"
P_WIDTH = 900
P_HEIGHT = 500

# tooltips = [("Hora", "@time"), ("Potencia", "$y{0} W")]
# HTML_TROW = """<tr>
# <td style="font-size: 15px; font-weight: bold;">{}</td>
# <td style="font-size: 15px; font-weight: bold; color: $color[hex];">{}</td>
# </tr>"""
HTML_TROW = """<tr><td style="font-size: 15px; font-weight: bold;">
<span class="bk-tooltip-color-block" style="background-color: {2}"></span>{0}</td>
<td style="font-size: 16px; font-weight: bold; color: {2};">{1}</td></tr>"""
TOOLTIP_ROWS = """<tr><td style="font-size: 15px; font-weight: bold;">Time:</td>
<td style="font-size: 15px; font-weight: bold;">@time</td></tr>"""


def _append_hover(df, delta_min=0, multi_day=False):
    if multi_day:
        df['time'] = [(x + pd.Timedelta(minutes=delta_min)).strftime("%-d/%b %H:%M:%S") for x in df.index]
    else:
        df['time'] = [(x + pd.Timedelta(minutes=delta_min)).strftime("%H:%M:%S.%f")[:-3] for x in df.index]
    return df


def _titulo_rango_temporal(extremos_dia):
    n_days = (extremos_dia[1] - extremos_dia[0]).days
    if n_days > 1:
        return "{:%-d de %B'%y} --> {:%-d de %B'%y}".format(extremos_dia[0], extremos_dia[1])
    else:
        return '{:%A %-d de %B de %Y}'.format(extremos_dia[0])


def _get_extremos_dia(df):
    extremos = [x.replace(microsecond=0) for x in df.index[[0, -1]]]
    # extremos_dia = [pd.Timestamp(x.date() + pd.Timedelta(days=x.hour // 12)) for x in extremos]
    # return [extremos[0].date(), (extremos[1] - pd.Timedelta(minutes=1) + pd.Timedelta(days=1)).date()]
    return extremos


# def _append_boxes(p, levels=(500, 3000), horiz_box=True, alpha=0.05, axis=None, is_band=False):
#     kwargs_box = dict(plot=p, fill_alpha=alpha)
#     if axis is not None:
#         kwargs_box.update(y_range_name=axis)
#     boxes = []
#     if is_band:
#         if horiz_box:
#             boxes.append(BoxAnnotation(top=levels[0], fill_color='red', **kwargs_box))
#             boxes.append(BoxAnnotation(bottom=levels[0], top=levels[1], fill_color='orange', **kwargs_box))
#             boxes.append(BoxAnnotation(bottom=levels[1], top=levels[2], fill_color='green', **kwargs_box))
#             boxes.append(BoxAnnotation(bottom=levels[2], top=levels[3], fill_color='orange', **kwargs_box))
#             boxes.append(BoxAnnotation(bottom=levels[3], fill_color='red', **kwargs_box))
#         else:
#             boxes.append(BoxAnnotation(right=levels[0], fill_color='red', **kwargs_box))
#             boxes.append(BoxAnnotation(left=levels[0], right=levels[1], fill_color='orange', **kwargs_box))
#             boxes.append(BoxAnnotation(left=levels[1], right=levels[2], fill_color='green', **kwargs_box))
#             boxes.append(BoxAnnotation(left=levels[2], right=levels[3], fill_color='orange', **kwargs_box))
#             boxes.append(BoxAnnotation(left=levels[3], fill_color='red', **kwargs_box))
#     else:
#         if horiz_box:
#             boxes.append(BoxAnnotation(top=levels[0], fill_color='green', **kwargs_box))
#             boxes.append(BoxAnnotation(bottom=levels[0], top=levels[1], fill_color='orange', **kwargs_box))
#             boxes.append(BoxAnnotation(bottom=levels[1], fill_color='red', **kwargs_box))
#         else:
#             boxes.append(BoxAnnotation(right=levels[0], fill_color='green', **kwargs_box))
#             boxes.append(BoxAnnotation(left=levels[0], right=levels[1], fill_color='orange', **kwargs_box))
#             boxes.append(BoxAnnotation(left=levels[1], fill_color='red', **kwargs_box))
#     p.renderers.extend(boxes)
#

def get_bokeh_version():
    """
    For templates (css & js src's)
    """
    return bokeh.__version__


def _return_html_comps(plots):
    script, divs = components(plots)
    reset_output()
    return script, divs, get_bokeh_version()


# def _get_axis_boxes_conf(axis):
#     levels = (500, 3000)
#     return {'axis': axis, 'levels': levels, 'horiz_box': True, 'alpha': 0.05, 'is_band': len(levels) > 2}


def _get_figure_plot(extremos, y_range, **fig_kwargs):
    kwargs = dict(x_range=extremos, y_range=y_range,
                  tools=TOOLS, active_drag=None, plot_width=P_WIDTH, plot_height=P_HEIGHT,
                  x_axis_type="datetime", toolbar_location="right", toolbar_sticky=False,  # "above",
                  title=_titulo_rango_temporal(extremos), responsive=True, **fig_kwargs)
    return figure(**kwargs)


def _format_axis_plot(p, color, label, unit):
    # Axis formatting
    p.axis.major_tick_in = 7
    p.axis.minor_tick_in = 3
    p.axis.major_tick_out = p.axis.minor_tick_out = 0
    p.title.text_font_size = '12pt'
    if unit:
        p.yaxis.axis_label = '{} ({})'.format(label, unit)
    else:
        p.yaxis.axis_label = label
    p.yaxis.axis_label_text_font_size = "11pt"
    p.yaxis.axis_line_color = p.yaxis.major_label_text_color = p.yaxis.axis_label_text_color = color
    p.yaxis.major_tick_line_color = p.yaxis.minor_tick_line_color = color


def _format_legend_plot(p):
    # Legend formatting
    p.legend.location = "top_left"
    p.legend.label_text_font_size = "7pt"
    p.legend.background_fill_alpha = .5
    p.legend.label_height = 12
    p.legend.glyph_height = 10
    p.legend.label_standoff = 5
    p.legend.padding = 8
    p.legend.spacing = 3


def _plot_bokeh_multi_index(data_plot, **fig_kwargs):
    # Bokeh does not work very well!! with timezones:
    data_plot = data_plot.tz_localize(None)

    y_range = [0, max(500, int(np.ceil(data_plot[COLS_DATA[0]].max().max() / ROUND_W) * ROUND_W))]
    minmax_ejes = [y_range, [0, 100]]

    extremos = _get_extremos_dia(data_plot)
    ejes = COLS_DATA[:2]

    # Figure
    p = _get_figure_plot(extremos, minmax_ejes[0], **fig_kwargs)

    tooltip_rows = TOOLTIP_ROWS
    for c, color, unit, label, fmt in zip(COLS_DATA, COLORS_DATA, UNITS_DATA, LABELS_DATA, FMT_TOOLTIP_DATA):
        tooltip_rows += HTML_TROW.format('{}:'.format(label), '@{}{} {}'.format(c, fmt, unit), color)
    p.add_tools(HoverTool(tooltips='<div><table>{}</table></div>'.format(tooltip_rows)))

    # Axis formatting
    _format_axis_plot(p, COLORS_DATA[0], LABELS_DATA[0], UNITS_DATA[0])
    if len(ejes) > 1:
        positions = (['right', 'left'] * 2)[:len(ejes[1:])]
        for extra_eje, pos, minmax, color, label, unit in zip(ejes[1:], positions, minmax_ejes[1:],
                                                              COLORS_DATA[1:], LABELS_DATA[1:], UNITS_DATA[1:]):
            p.extra_y_ranges[extra_eje] = Range1d(*minmax)
            axkw = dict(y_range_name=extra_eje,
                        axis_label='{} ({})'.format(label, unit),
                        axis_label_text_font_size="10pt", axis_line_color=color,
                        major_label_text_color=color, axis_label_text_color=color,
                        major_tick_line_color=color, minor_tick_line_color=color,
                        major_tick_in=5, major_tick_out=0, minor_tick_in=3, minor_tick_out=0)
            p.add_layout(LinearAxis(**axkw), pos)

    # Make data source w/ time hover
    data = ColumnDataSource(_append_hover(data_plot.round(2), multi_day=(extremos[1] - extremos[0]).days > 1))

    # Plot lines
    # kwargs_p = dict(source=data, alpha=.9, line_width=2, legend=LABELS_DATA[0], color=color_base)
    # kwargs_l = dict(source=data, alpha=.9, line_width=2, legend=LABELS_DATA[1], color=COLORS_DATA[1],
    kwargs_p = dict(source=data, alpha=.9, line_width=2, color=COLORS_DATA[0])
    kwargs_l = dict(source=data, alpha=.9, line_width=2, color=COLORS_DATA[1], y_range_name=COLS_DATA[1])
    p.line('ts', 'power', **kwargs_p)
    p.line('ts', 'ldr', **kwargs_l)

    # Plot patch
    zero = np.array([0])
    df_patch = pd.Series(data_plot['ldr'].fillna(method='pad', limit=2).fillna(0).round(2))
    x = np.append(np.insert(df_patch.index.values, 0, df_patch.index.values[0]), df_patch.index.values[-1])
    y = np.append(np.insert(df_patch.values, 0, zero), zero)
    kwargs_patch = dict(color=COLORS_DATA[1], line_alpha=0, fill_alpha=0.10, y_range_name=COLS_DATA[1])
    p.patch(x, y, **kwargs_patch)

    df_patch = data_plot['power'].fillna(method='pad', limit=2).fillna(0).round(2)
    x = np.append(np.insert(df_patch.index.values, 0, df_patch.index.values[0]), df_patch.index.values[-1])
    y = np.append(np.insert(df_patch.values, 0, zero), zero)
    kwargs_patch = dict(color=COLORS_DATA[0], line_alpha=0, fill_alpha=0.15)
    p.patch(x, y, **kwargs_patch)

    # Legend formatting
    # _format_legend_plot(p)
    return p


def _plot_bokeh_hourly(data_plot, **fig_kwargs):
    # Bokeh does not work very well!! with timezones:
    data_plot = data_plot.tz_localize(None)

    data_plot['ts_mid'] = data_plot.index + pd.Timedelta('30min')
    data_plot.index += pd.Timedelta('30min')

    y_range = [0, max(1, np.ceil(data_plot[COLS_DATA_KWH[0]].max().max() / ROUND_KWH) * ROUND_KWH)]
    minmax_ejes = [y_range, [0, max(500, int(np.ceil(data_plot[COLS_DATA_KWH[1]].max().max() / ROUND_W) * ROUND_W))]]
    extremos = _get_extremos_dia(data_plot)
    ejes = COLS_DATA_KWH[:2]

    # Figure
    p = _get_figure_plot(extremos, minmax_ejes[0], **fig_kwargs)

    tooltip_rows = TOOLTIP_ROWS
    for c, color, unit, label, fmt in zip(COLS_DATA_KWH, COLORS_DATA_KWH,
                                          UNITS_DATA_KWH, LABELS_DATA_KWH, FMT_TOOLTIP_DATA_KWH):
        tooltip_rows += HTML_TROW.format('{}:'.format(label), '@{}{} {}'.format(c, fmt, unit), color)
    p.add_tools(HoverTool(tooltips='<div><table>{}</table></div>'.format(tooltip_rows)))

    # Axis formatting
    _format_axis_plot(p, COLORS_DATA_KWH[0], LABELS_DATA_KWH[0], UNITS_DATA_KWH[0])
    if len(ejes) > 1:
        positions = (['right', 'left'] * 2)[:len(ejes[1:])]
        for extra_eje, pos, minmax, color, label, unit in zip(ejes[1:], positions, minmax_ejes[1:], COLORS_DATA_KWH[1:],
                                                              LABELS_DATA_KWH[1:], UNITS_DATA_KWH[1:]):
            p.extra_y_ranges[extra_eje] = Range1d(*minmax)
            axkw = dict(y_range_name=extra_eje,
                        axis_label='{} ({})'.format(label, unit),
                        axis_label_text_font_size="10pt", axis_line_color=color,
                        major_label_text_color=color, axis_label_text_color=color,
                        major_tick_line_color=color, minor_tick_line_color=color,
                        major_tick_in=5, major_tick_out=0, minor_tick_in=3, minor_tick_out=0)
            p.add_layout(LinearAxis(**axkw), pos)

    # Make data source w/ time hover
    data = ColumnDataSource(_append_hover(data_plot, delta_min=-30, multi_day=(extremos[1] - extremos[0]).days > 1))

    # Plot lines
    kwargs_kwh = dict(source=data, x='ts', width=3600000, bottom=0, top=COLS_DATA_KWH[0], legend=LABELS_DATA_KWH[0],
                      fill_alpha=.7, line_alpha=.9, line_width=1, line_join='round', color=COLORS_DATA_KWH[0])
    p.vbar(**kwargs_kwh)

    kwargs_l = dict(source=data, alpha=.95, line_width=1.5, line_join='bevel', color=COLORS_DATA_KWH[1],
                    y_range_name=COLS_DATA_KWH[1], legend=LABELS_DATA_KWH[1])
    p.line('ts', COLS_DATA_KWH[1], **kwargs_l)
    kwargs_l.update(color=COLORS_DATA_KWH[2], legend=LABELS_DATA_KWH[2])
    p.line('ts', COLS_DATA_KWH[2], **kwargs_l)

    # # Plot patch
    # df_patch = pd.Series(data_plot['ldr'].fillna(method='pad', limit=2).fillna(0).round(2))
    # x = np.append(np.insert(df_patch.index.values, 0, df_patch.index.values[0]), df_patch.index.values[-1])
    # y = np.append(np.insert(df_patch.values, 0, 0), 0)
    # kwargs_patch = dict(color=COLORS_DATA[1], line_alpha=0, fill_alpha=0.10, y_range_name=COLS_DATA[1])
    # p.patch(x, y, **kwargs_patch)
    #
    # df_patch = data_plot['power'].fillna(method='pad', limit=2).fillna(0).round(2)
    # x = np.append(np.insert(df_patch.index.values, 0, df_patch.index.values[0]), df_patch.index.values[-1])
    # y = np.append(np.insert(df_patch.values, 0, 0), 0)
    # kwargs_patch = dict(color=color_base, line_alpha=0, fill_alpha=0.15)
    # p.patch(x, y, **kwargs_patch)

    # Legend formatting
    _format_legend_plot(p)
    return p


def html_plot_buffer_bokeh(data_plot, is_kwh_plot=False, **fig_kwargs):
    """
    Given a pandas DataFrame (or a list of df's), returns the html components for rendering the graph.
    :return script, divs, bokeh.__version__
    """
    if is_kwh_plot:
        return _return_html_comps([_plot_bokeh_hourly(data_plot, **fig_kwargs)])
    else:
        return _return_html_comps([_plot_bokeh_multi_index(data_plot, **fig_kwargs)])



# -*- coding: utf-8 -*-
from collections import OrderedDict
from termcolor import cprint


# Colorized output:
def print_info(x):
    """Prints in blue"""
    cprint(x, 'blue')


def print_infob(x):
    """Prints in bold + blue"""
    cprint(x, 'blue', attrs=['bold'])


def print_ok(x):
    """Prints in bold + green"""
    cprint(x, 'green', attrs=['bold'])


def print_secc(x):
    """Prints in bold + blue + underline & starts with ' ==> '"""
    cprint(' ==> ' + x, 'blue', attrs=['bold', 'underline'])


def print_err(x):
    """Prints in bold + red background & starts with 'ERROR: '"""
    cprint('ERROR: ' + str(x), on_color='on_red', attrs=['bold'])


def print_warn(x):
    """Prints in magenta & starts with 'WARNING: '"""
    cprint('WARNING: ' + str(x), 'magenta')


def print_bold(x):
    """Prints in bold"""
    cprint(x, attrs=['bold'])


def print_boldu(x):
    """Prints in bold + underline"""
    cprint(x, 'grey', attrs=['bold', 'underline'])


def print_yellowb(x):
    """Prints in yellow + underline"""
    cprint(x, 'yellow', attrs=['bold'])


def print_grey(x):
    """Prints in grey"""
    cprint(x, 'grey')


def print_greyb(x):
    """Prints in bold + grey"""
    cprint(x, 'grey', attrs=['bold'])


def print_red(x):
    """Prints in red"""
    cprint(x, 'red')


def print_redb(x):
    """Prints in bold + red"""
    cprint(x, 'red', attrs=['bold'])  # en shared


def print_green(x):
    """Prints in green"""
    cprint(x, 'green')


def print_yellow(x):
    """Prints in yellow"""
    cprint(x, 'yellow')


def print_blue(x):
    """Prints in blue"""
    cprint(x, 'blue')


def print_magenta(x):
    """Prints in magenta"""
    cprint(x, 'magenta')


def print_cyan(x):
    """Prints in cyan"""
    cprint(x, 'cyan')


def print_white(x):
    """Prints in white"""
    cprint(x, 'white')


def ppdict(dict_to_print, br='\n', html=False, key_align='l', sort_keys=True,
           key_preffix='', key_suffix='', value_prefix='', value_suffix='', left_margin=3, indent=2):
    """Indent representation of a dict"""
    if dict_to_print:
        if sort_keys:
            dic = dict_to_print.copy()
            keys = list(dic.keys())
            keys.sort()
            dict_to_print = OrderedDict()
            for k in keys:
                dict_to_print[k] = dic[k]

        tmp = ['{']
        ks = [type(x) == str and "'%s'" % x or x for x in dict_to_print.keys()]
        vs = [type(x) == str and "'%s'" % x or x for x in dict_to_print.values()]
        max_key_len = max([len(str(x)) for x in ks])

        for i in range(len(ks)):
            k = {1: str(ks[i]).ljust(max_key_len),
                 key_align == 'r': str(ks[i]).rjust(max_key_len)}[1]

            v = vs[i]
            tmp.append(' ' * indent + '{}{}{}:{}{}{},'.format(key_preffix, k, key_suffix,
                                                              value_prefix, v, value_suffix))

        tmp[-1] = tmp[-1][:-1]  # remove the ',' in the last item
        tmp.append('}')

        if left_margin:
            tmp = [' ' * left_margin + x for x in tmp]

        if html:
            return '<code>{}</code>'.format(br.join(tmp).replace(' ', '&nbsp;'))
        else:
            return br.join(tmp)
    else:
        return '{}'



# -*- coding: utf-8 -*-
from flask import Flask
from werkzeug.contrib.fixers import ProxyFix
from werkzeug.routing import Rule
import jinja2
import os
from enerpi.base import CONFIG, DATA_PATH, check_resource_files


basedir = os.path.abspath(os.path.dirname(__file__))
STATIC_PATH = os.path.join(DATA_PATH, CONFIG.get('ENERPI_WEBSERVER', 'STATIC_PATH', fallback='WWW'))
LOGGING_LEVEL_SERVER = CONFIG.get('ENERPI_WEBSERVER', 'LOGGING_LEVEL_WEB', fallback='DEBUG')
SERVER_FILE_LOGGING = os.path.join(STATIC_PATH,
                                   CONFIG.get('ENERPI_WEBSERVER', 'FILE_LOGGING_WEB', fallback='enerpiweb.log'))
PREFIX_WEB = CONFIG.get('ENERPI_WEBSERVER', 'PREFIX_WEB', fallback='/enerpi')
BASECOLOR = '#{}'.format(CONFIG.get('ENERPI_WEBSERVER', 'BASECOLOR_HEX', fallback='0CBB43'))
check_resource_files(STATIC_PATH, os.path.join(basedir, 'static'))

# WITH_WEB = CONFIG.getboolean('ENERPI_WEBSERVER', 'WITH_WEBSERVER', fallback=True)
WITH_ML_SUBSYSTEM = CONFIG.getboolean('ENERPI_WEBSERVER', 'WITH_ML', fallback=False)

# FLASK APP
app = Flask(__name__, static_path=PREFIX_WEB + '/static', static_folder=STATIC_PATH)
app.url_rule_class = lambda path, **options: Rule(PREFIX_WEB + path, **options)
app.jinja_env.trim_blocks = True
app.jinja_env.lstrip_blocks = True
app.jinja_env.cache = {}
app.config['PREFIX_WEB'] = PREFIX_WEB
app.config['BASECOLOR'] = BASECOLOR
app.config['WITH_ML_SUBSYSTEM'] = WITH_ML_SUBSYSTEM
app.jinja_loader = jinja2.FileSystemLoader(os.path.join(basedir, 'templates'))

# Views
# noinspection PyUnresolvedReferences,PyPep8
from enerpiweb import views, views_filehandler, utils, rt_stream

# wsgi
app.wsgi_app = ProxyFix(app.wsgi_app)



# -*- coding: utf-8 -*-
from cryptography.fernet import Fernet, InvalidToken
from binascii import Error
import os
import socket
import sys
from time import time
from enerpi.base import DATA_PATH, CONFIG, log


# LAN broadcasting
UDP_IP = CONFIG.get('BROADCAST', 'UDP_IP', fallback="192.168.1.255")
UDP_PORT = CONFIG.getint('BROADCAST', 'UDP_PORT', fallback=57775)
DESCRIPTION_IO = "\tSENDER - RECEIVER vía UDP. Broadcast IP: {}, PORT: {}".format(UDP_IP, UDP_PORT)
KEY_FILE = os.path.join(DATA_PATH, CONFIG.get('BROADCAST', 'KEY_FILE', fallback='.secret_key'))


# TODO Automatize get encryption key (for tests)?
def get_encryption_key(key_file):
    """
    Encryption key for symmetric encryption. URL-safe base64-encoded 32-byte key

    * Get local key, or exec wizard for making one.

    :param: key_file: :str: path of the file with the encryption key
    :return: :bytes: encryption_key
    """

    try:
        secret_key = open(key_file).read().encode()
    except FileNotFoundError:
        print('\033[0m\033[1m\033[34mCrypto KEY is missing.\033[0m Please, select one option:\n'
              '\t\033[1m1. Generate a new key.\n\t2. Input an existent key. (Copy&paste from another computer)')
        selec = input('\033[34mOPTION: \033[0m\033[31m')
        if selec == '1':
            secret_key = Fernet.generate_key()
        else:
            secret_key = input('*** KEY: ').encode()
        if len(secret_key) > 10:
            print('\033[0mThe new KEY ("{}")\nwill be saved in "{}"\n'.format(secret_key, key_file))
            open(key_file, 'wb').write(secret_key)
        else:
            print('\033[31;1mNot a valid KEY!:\n"{}". Try again... BYE!\033[0m'.format(secret_key))
            sys.exit(-1)
    return secret_key


def get_codec(secret_key):
    """
    Get Fernet CODEC from secret_key with an AssertionError if Fernet error raises.

    :param secret_key: :bytes: encryption_key
    :return: Fernet()

    """
    # secret_key = get_encryption_key(key_file)
    try:
        codec = Fernet(secret_key)
        return codec
    except Error as error_fernet:
        log('Crypto KEY is not a valid KEY! -> {}.\nKEY="{}". Try again... BYE!'
            .format(error_fernet, secret_key), 'error', True)
        assert 0, 'Crypto KEY is not a valid KEY: {}'.format(error_fernet)
        # sys.exit(-1)


CODEC = get_codec(get_encryption_key(KEY_FILE))


def receiver_msg_generator(verbose=True, n_msgs=None, port=UDP_PORT, codec=CODEC):
    """
    Generador de mensajes en el receptor de la emisión en la broadcast IP. Recibe los mensajes y los desencripta.
    También devuelve los tiempos implicados en la recepción (~ el ∆T entre mensajes) y el proceso de desencriptado.
    :param verbose: :bool: Imprime Broadcast IP & PORT.
    :param n_msgs: :int: # de mensajes a recibir (ilimitados por defecto).
    :param port: :int: # de puerto de escucha.
    :param codec: :Fernet obj: Fernet object for decrypting msgs.
    :yield: msg, ∆T_msg, ∆T_decryp
    """
    sock, counter = None, 0
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.bind((UDP_IP, port))
        log(DESCRIPTION_IO, 'ok', verbose, False)
        while (n_msgs is None) or (counter < n_msgs):
            tic = time()
            data, addr = sock.recvfrom(1024)  # buffer size is 1024 bytes
            toc_msg = time()
            try:
                msg = codec.decrypt(data).decode()
            except InvalidToken as e:
                log('InvalidToken ERROR: {}. La clave es correcta?\n* KEY: "{}"'
                    .format(e, get_encryption_key(key_file=KEY_FILE)), 'error', verbose, True)
                break
            toc_dcry = time()
            yield msg, toc_msg - tic, toc_dcry - toc_msg
            counter += 1
        # log('Closing receiver_msg_generator socket...', 'debug', verbose, True)
    # except OSError as e:
    #     log('OSError {} en receiver_msg_generator'.format(e), 'error', verbose, True)
    except KeyboardInterrupt:
        log('KeyboardInterrupt en receiver_msg_generator', 'warn', verbose, True)
    if sock is not None:
        sock.close()
    raise StopIteration


def broadcast_msg(msg, counter_unreachable, sock_send=None, codec=CODEC, verbose=True):
    """
    Emisión de datos en modo broadcast UDP (para múltiples receptores) como mensaje de texto encriptado.
    :param msg: Cadena de texto a enviar.
    :param counter_unreachable: np.array([0, 0]) de control de emisiones incorrectas (seguidas y totales)
    :param sock_send: Socket de envío broadcast. Se devuelve para su reutilización.
    :param codec: :Fernet obj: Fernet object for encrypting msgs.
    :param verbose: Imprime en stout mensajes de error y de envío de datos
    :return: sock_send
    """

    def _get_broadcast_socket(verb=False):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        # sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 32)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        log(DESCRIPTION_IO, 'ok', verb, verb)
        return sock

    if sock_send is None:
        sock_send = _get_broadcast_socket()
    encrypted_msg_b = codec.encrypt(msg.encode())
    try:
        sock_send.sendto(encrypted_msg_b, (UDP_IP, UDP_PORT))
        counter_unreachable[0] = 0
    except OSError as e:  # [Errno 101] Network is unreachable
        log('OSError: {}; C_UNREACHABLE: {}'.format(e, counter_unreachable), 'warn', verbose)
        counter_unreachable += 1
        sock_send = None
    # except Exception as e:
    #     log('ERROR en sendto: {} [{}]'.format(e, e.__class__), 'err', verbose)
    #     sock_send = _get_broadcast_socket()
    #     sock_send.sendto(encrypted_msg_b, (UDP_IP, UDP_PORT))
    log('SENDED: {}'.format(msg), 'debug', verbose, False)
    return sock_send



# -*- coding: utf-8 -*-
from collections import OrderedDict
import configparser
import locale
import logging
import os
import pytz
import shutil
import subprocess
import sys
import textwrap
from threading import Timer
from time import time, sleep
from enerpi import BASE_PATH, PRETTY_NAME
from enerpi.prettyprinting import print_err, print_red, print_info, print_ok, print_warn, print_yellowb, print_magenta
from enerpi.pitemps import get_cpu_temp, get_gpu_temp


ENCODING = 'UTF-8'
CONFIG_FILENAME = 'config_enerpi.ini'


class EnerpiAnalogSensor(object):
    """
    Object for centralize access to sensor properties, as description, color, units, etc.

    """
    def __init__(self, column_name, description='', channel=0, channel_name='CH_PROBE',
                 is_rms=True, color='0CBB43', bias=0):
        self.name = column_name
        self.description = description
        self.channel = channel
        self.channel_name = channel_name
        self.is_rms = is_rms
        self.unit = 'W' if is_rms else '%'
        self.color = '#{}'.format(color)
        self.bias = bias

    def __repr__(self):
        return 'ANALOG_S: {} ({}{}, ch={}, unit={})'.format(self.name, 'RMS, ' if self.is_rms else '',
                                                            self.description, self.channel, self.unit)


class EnerpiSamplerConf(object):
    """
    Object for easy access to data properties, as # of columns, column names, descriptions and units, type of sensors,
    color for plotting, summary info & columns, etc...

    """
    def __init__(self, configparser_obj):
        # Read config INI file
        self._fmt_ts = configparser_obj.get('ENERPI_SAMPLER', 'FMT_TS', fallback='%Y-%m-%d %H:%M:%S.%f')
        self._col_ts = configparser_obj.get('ENERPI_SAMPLER', 'COL_TS', fallback='ts')
        self._ref_column_rms = 'ref'
        self._ref_column_mean = 'ref_n'

        # Voltage divisor for sct030 probes:
        mcp3008_dac_prec = 10  # bits
        niveles = 2 ** mcp3008_dac_prec - 1
        bias_current = -(niveles // 2) / niveles

        # Get columns
        cols_data_rms = configparser_obj.get('ENERPI_SAMPLER', 'COLS_DATA_RMS', fallback='power').split(', ')
        cols_data_mean = configparser_obj.get('ENERPI_SAMPLER', 'COLS_DATA_MEAN', fallback='noise, ldr').split(', ')

        # Get descriptions
        descr_data_rms = configparser_obj.get('ENERPI_SAMPLER', 'COLS_DATA_RMS', fallback='power').split(', ')
        descr_data_mean = configparser_obj.get('ENERPI_SAMPLER', 'COLS_DATA_MEAN', fallback='noise, ldr').split(', ')

        # Get colors
        colors_data_rms = configparser_obj.get('ENERPI_SAMPLER', 'COLORS_DATA_RMS', fallback='power').split(', ')
        colors_data_mean = configparser_obj.get('ENERPI_SAMPLER', 'COLORS_DATA_MEAN', fallback='noise, ldr').split(', ')
        # Get channel probes
        names_channels = list(configparser_obj['MCP3008'].keys())
        channels = [configparser_obj.getint('MCP3008', ch_name, fallback=i) for i, ch_name in enumerate(names_channels)]

        self._sensors_rms = [EnerpiAnalogSensor(name, desc, ch, ch_n, True, color, bias_current)
                             for name, desc, color, ch, ch_n in zip(cols_data_rms, descr_data_rms, colors_data_rms,
                                                                    channels[:len(cols_data_rms)],
                                                                    names_channels[:len(cols_data_rms)]) if ch >= 0]
        self._sensors_mean = [EnerpiAnalogSensor(name, desc, ch, ch_n, False, color)
                              for name, desc, color, ch, ch_n in zip(cols_data_mean, descr_data_mean, colors_data_mean,
                                                                     channels[len(cols_data_rms):],
                                                                     names_channels[len(cols_data_rms):]) if ch >= 0]
        # PiSampler parameters:
        # Current meter
        # Voltaje típico RMS de la instalación a medir. (SÓLO SE ESTIMA P_ACTIVA!!)
        voltaje = CONFIG.getint('ENERPI_SAMPLER', 'VOLTAJE', fallback=236)
        # 30 A para 1 V --> Pinza amperométrica SCT030-030
        a_ref = CONFIG.getfloat('ENERPI_SAMPLER', 'A_REF', fallback=30.)
        # V, V_ref RPI GPIO
        v_ref = CONFIG.getfloat('ENERPI_SAMPLER', 'V_REF', fallback=3.3)
        self.voltaje = voltaje
        self.rms_multiplier = self.voltaje * a_ref * v_ref

        # Sampling:
        self.ts_data_ms = CONFIG.getint('ENERPI_SAMPLER', 'TS_DATA_MS', fallback=12)
        # ∆T para el deque donde se acumulan frames
        self.rms_roll_window_sec = CONFIG.getfloat('ENERPI_SAMPLER', 'RMS_ROLL_WINDOW_SEC', fallback=2.)
        s_calc = self.ts_data_ms if self.ts_data_ms > 0 else 8
        self.n_samples_buffer_rms = int(round(self.rms_roll_window_sec * 1000 / s_calc))

        self.delta_sec_data = CONFIG.getfloat('ENERPI_SAMPLER', 'DELTA_SEC_DATA', fallback=1.)
        self.measure_ldr_divisor = CONFIG.getint('ENERPI_SAMPLER', 'MEASURE_LDR_DIVISOR', fallback=10)
        self.n_samples_buffer_mean = self.n_samples_buffer_rms // self.measure_ldr_divisor
        self.TZ = pytz.timezone(CONFIG.get('ENERPI_SAMPLER', 'TZ', fallback='Europe/Madrid'))

    def __repr__(self):
        repr_str = 'ENERPI SAMPLER SENSORS:\n * '
        repr_str += '\n * '.join([str(s) for s in self])
        repr_str += ('\n --SAMPLING: V={} V; ∆T={} s; sampling={} ms, N={}, {}(/{})'
                     .format(self.voltaje, self.delta_sec_data, self.ts_data_ms,
                             self.n_samples_buffer_rms, self.n_samples_buffer_mean, self.measure_ldr_divisor))
        return repr_str

    def __len__(self):
        return len(self._sensors_rms) + len(self._sensors_mean)

    def __getitem__(self, key):
        if type(key) is int:
            if key >= len(self._sensors_rms):
                try:
                    return self._sensors_mean[key - len(self._sensors_rms)]
                except IndexError:
                    raise KeyError('Sensor #{} not present (available sensors: {})'
                                   .format(key, self.columns_sensors[1:]))
            return self._sensors_rms[key]
        else:
            assert type(key) is str
            try:
                int_key = self.columns_sensors.index(key) - 1
                return self.__getitem__(int_key)
            except ValueError:
                raise KeyError("Sensor '{}' not present (available sensors: {})".format(key, self.columns_sensors[1:]))

    def __iter__(self):
        for x in self._sensors_rms + self._sensors_mean:
            yield x

    @property
    def main_column(self):
        """
        First (& principal) RMS sensor (Main power)

        :return: :str: name of main column
        """
        return self._sensors_rms[0].name

    @property
    def ts_column(self):
        """
        Name of time-series index

        :return: :str: name of time-series index column
        """
        return self._col_ts

    @property
    def ts_fmt(self):
        """
        String formatting of timestamps:

        :return: :str: fmt
        """
        return self._fmt_ts

    @property
    def ref_rms(self):
        """
        Name of column with # of samples of RMS values

        :return: :str: column name
        """
        return self._ref_column_rms

    @property
    def ref_mean(self):
        """
        Name of column with # of samples of MEAN values

        :return: :str: column name
        """
        return self._ref_column_mean

    @property
    def n_cols_sensors(self):
        """
        # of columns: index column (TS) + RMS sensors + MEAN sensors

        :return: :int: # of columns
        """
        return len(self) + 1

    @property
    def n_cols_sampling(self):
        """
        # of columns: index column (TS) + RMS sensors + MEAN sensors + ref_RMS + ref_MEAN

        :return: :int: # of columns
        """
        return len(self) + 3

    @property
    def columns_sensors(self):
        """
        List of columns names of sensors (+ts): index column (TS) + RMS sensors + MEAN sensors

        :return: :list:
        """
        return [self.ts_column] + [s.name for s in self._sensors_rms + self._sensors_mean]

    @property
    def columns_sensors_rms(self):
        """
        List of columns names of RMS sensors

        :return: :list:
        """
        return [s.name for s in self._sensors_rms]

    @property
    def columns_sensors_mean(self):
        """
        List of columns names of MEAN sensors

        :return: :list:
        """
        return [s.name for s in self._sensors_mean]

    @property
    def columns_sampling(self):
        """
        List of columns names of sampling data index column (TS) + RMS sensors + MEAN sensors + REF_RMS + REF_MEAN

        :return: :list:
        """
        return self.columns_sensors + [self.ref_rms, self.ref_mean]

    def included_columns_sensors(self, dict_sample):
        """
        Return RMS & MEAN columns included in dict sample

        :return: :tuple: tuple with 2 lists (rms & mean included columns)

        """
        cols_rms = [s.name for s in self._sensors_rms if s.name in dict_sample]
        cols_mean = [s.name for s in self._sensors_mean if s.name in dict_sample]
        return cols_rms, cols_mean

    def included_columns_sampling(self, dict_sample):
        """
        Return RMS & MEAN columns (or ref_RMS and ref_MEAN) included in sampling data dict

        :return: :tuple: tuple with 3 lists (rms, mean & ref included columns)

        """
        cols_rms, cols_mean = self.included_columns_sensors(dict_sample)
        cols_ref = [c for c in [self._ref_column_rms, self._ref_column_mean] if c in dict_sample]
        return cols_rms, cols_mean, cols_ref

    def descriptions(self, columns, return_list=True):
        """
        Return column descriptions (as list or as dict c:desc)
        :param columns: list of columns
        :param return_list: :bool: False for dict return
        :return: :list: or :dict:
        """
        def _get_desc(key):
            try:
                return self[key].description
            except KeyError as e:
                if key == self.ts_column:
                    return 'TS'
                elif key == self.ref_rms:
                    return '# samples'
                elif key == self.ref_mean:
                    return '# samples_n'
                else:
                    raise KeyError(e)

        if return_list:
            return [_get_desc(c) for c in columns]
        else:
            return {c: _get_desc(c) for c in columns}


def _makedirs(dest_path):
    try:
        os.makedirs(os.path.dirname(dest_path), exist_ok=True)
        if os.path.exists(dest_path):
            os.utime(dest_path, None)
    except PermissionError as e:
        log('PermissionError: {}'.format(e), 'error', True, False)
        # sudo chmod 777 ~/ENERPIDATA/enerpi.log
        # TODO Notify error
        sys.exit(2)


def _funcs_tipo_output(tipo_log):
    """
    Functions for printing and logging based on type.

    :param tipo_log: :enum: error, debug, ok, info, warn, magenta
    :return: print_func, logging_func
    """
    if tipo_log == 'error':
        return print_err, logging.error
    elif tipo_log == 'debug':
        return print_red, logging.debug
    elif tipo_log == 'ok':
        return print_ok, logging.info
    elif tipo_log == 'info':
        return print_info, logging.info
    elif (tipo_log == 'warn') or (tipo_log == 'warning'):
        return print_warn, logging.warning
    elif tipo_log == 'magenta':
        return print_magenta, logging.warning
    else:
        return print_yellowb, logging.debug


def log(msg, tipo, verbose=True, log_msg=True):
    """
    Logging wrapper to log and / or color-print

    :param msg:
    :param tipo:
    :param verbose: :bool:
    :param log_msg: :bool:
    """
    f1, f2 = _funcs_tipo_output(tipo)
    if verbose:
        f1(msg)
    if log_msg:
        f2(msg)


def set_logging_conf(filename, level='DEBUG', verbose=True, with_initial_log=True):
    """Logging configuration"""
    _makedirs(filename)
    # os.makedirs(os.path.dirname(filename), exist_ok=True)
    # try:
    logging.basicConfig(filename=filename, level=level, datefmt='%d/%m/%Y %H:%M:%S',
                        format='%(levelname)s [%(filename)s_%(funcName)s] - %(asctime)s: %(message)s')
    # except PermissionError as e:
    #     log('PermissionError: {}'.format(e), 'error', True, False)
    #     sudo chmod 777 ~/ENERPIDATA/enerpi.log
    #     sys.exit(2)
    if with_initial_log:
        log(PRETTY_NAME, 'ok', verbose)


def show_pi_temperature(log_rpi_temps, ts=3):
    """
    Sensor Raspberry PI temperatures, infinite-loop of logging / printing values

    :param log_rpi_temps: :bool: activate RPI temp logging
    :param ts: :int: ∆T between samples
    """
    def _loop_log_temps():
        while True:
            t_cpu = get_cpu_temp()
            t_gpu = get_gpu_temp()
            if (t_cpu is None) and (t_gpu is None):
                log('NO RPI TEMPS', 'warning', False, True)
                break
            log('Tªs --> {:.1f} / {:.1f} ºC'.format(t_cpu, t_gpu), 'otro', False, True)
            sleep(ts)

    if log_rpi_temps:
        # Shows RPI Temps
        timer_temps = Timer(.5, _loop_log_temps)
        timer_temps.start()
        return timer_temps
    return None


def timeit(cadena_log, verbose=False, *args_dec):
    """
    Decorator (wrapper) to timeit and log (and print) any function.

    (For debugging / optimize purposes)

    :param cadena_log:
    :param verbose:
    :param args_dec:
    :return:
    """
    def _real_deco(function):
        def _wrapper(*args, **kwargs):
            kwargs_print = {}
            tic = time()
            out = function(*args, **kwargs)
            if verbose:
                print_yellowb(cadena_log.format(*args_dec, **kwargs_print) + ' TOOK: {:.3f} s'.format(time() - tic))
            logging.debug(cadena_log.format(*args_dec, **kwargs_print) + ' TOOK: {:.3f} s'.format(time() - tic))
            return out
        return _wrapper
    return _real_deco


def get_lines_file(filename, tail=None, reverse=False):
    """
    Read (or tail) a text file.

    :param filename:
    :param tail:
    :param reverse:
    :return:
    """
    if os.path.exists(filename):
        if os.path.isfile(filename):
            try:
                if tail is not None:
                    output = subprocess.check_output(['/usr/bin/tail', '-n', '{}'.format(int(tail)), filename])
                    lines = output.decode().split('\n')
                else:
                    with open(filename, 'r', encoding=ENCODING) as file:
                        lines = file.read().split('\n')
                if len(lines) > 0 and lines[-1] == '':
                    lines = lines[:-1]
                if reverse:
                    return list(reversed(lines))
                return lines
            except Exception as e:
                return ['ERROR Reading {}: "{}" [{}]'.format(filename, e, e.__class__)]
        else:
            return ['{} is not a file!!'.format(filename)]
    return ['Path not found: "{}"'.format(filename)]


def config_dict_for_web_edit(lines_ini_file):
    """
    * Divide configuration INI file with this structure:
    [section]
    # comment
    ; comment
    VARIABLE = VALUE
    (discarding file header comments)

    Make Ordered dict like:
    ==> [(section_name,
            OrderedDict([(VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         ...
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment))])),
         (section_name,
            OrderedDict([(VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         ...
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment))])),
         ...]

    :param lines_ini_file: :list: Content of INI file after 'readlines()'
    :return: :OrderedDict:

    """
    def _is_bool_variable(value):
        value = value.lower()
        if (value == 'true') or (value == 'false'):
            return True, value == 'true'
        return False, None

    config_entries = OrderedDict()
    section, comment = None, None
    init = False
    for l in lines_ini_file:
        l = l.replace('\n', '').lstrip().rstrip()
        if l.startswith('['):
            section = l.replace('[', '').replace(']', '')
            init = True
            comment = None
            config_entries[section] = OrderedDict()
        elif l.startswith('#') or l.startswith(';'):
            if init:
                if comment is None:
                    comment = l[1:].lstrip()
                else:
                    comment += ' {}'.format(l[1:].lstrip())
        elif init and (len(l) > 0):
            # Read variable and append (/w comments)
            variable_name, variable_value = l.split('=')
            variable_name = variable_name.lstrip().rstrip()
            variable_value = variable_value.lstrip().rstrip()
            is_bool, bool_value = _is_bool_variable(variable_value)
            if not is_bool:
                try:
                    variable_value = int(variable_value)
                    var_type = 'int'
                except ValueError:
                    try:
                        variable_value = float(variable_value)
                        var_type = 'float'
                    except ValueError:
                        var_type = 'str'
            else:
                var_type = 'bool'
                variable_value = bool_value
            config_entries[section][variable_name] = variable_value, var_type, comment
            comment = None
    return config_entries


def config_changes(dict_web_form, dict_config):
    """
    Process changes in web editor

    :param dict_web_form: :OrderedDict: Posted Form values
    :param dict_config:  :OrderedDict: config dict of dicts (like the one 'config_dict_for_web_edit' returns)
    :return: :tuple: (:bool: changed, :list: updated variables, :OrderedDict: updated dict_config)

    """
    def _is_changed(value, params, name):
        if params[1] == 'int':
            try:
                value = int(value)
            except ValueError:
                value = float(value)
        elif params[1] == 'float':
            value = float(value)
        elif params[1] == 'bool':
            value = (value.lower() == 'true') or (value.lower() == 'on')
        if value != params[0]:
            log('"{}" -> HAY CAMBIO DE {} a {} (type={})'.format(name, params[0], value, params[1]), 'debug', False)
            return True, value
        return False, value

    dict_config_updated = dict_config.copy()
    dict_web_form = dict_web_form.copy()
    vars_updated = []
    for section, entries in dict_config_updated.items():
        for variable_name, variable_params in entries.items():
            if variable_name in dict_web_form:
                new_v = dict_web_form.pop(variable_name)
                changed, new_value = _is_changed(new_v, variable_params, variable_name)
                if changed:
                    vars_updated.append((variable_name, variable_params[0], new_value))
                    params_var = list(dict_config_updated[section][variable_name])
                    params_var[0] = new_value
                    dict_config_updated[section][variable_name] = tuple(params_var)
            elif (variable_params[1] == 'bool') and variable_params[0]:  # Bool en off en el form y True en config
                vars_updated.append((variable_name, variable_params[0], False))
                params_var = list(dict_config_updated[section][variable_name])
                params_var[0] = False
                log('"{}" -> HAY CHECKBOX CH DE {} a {} (type={})'
                    .format(variable_name, variable_params[0], False, variable_params[1]), 'debug', False)
                dict_config_updated[section][variable_name] = tuple(params_var)
    return len(vars_updated) > 0, vars_updated, dict_config_updated


def make_ini_file(dict_config, dest_path=None):
    """
    Makes INI file (and writes it if dest_path is not None) from an OrderedDict
    like the one 'config_dict_for_web_edit' returns.

    * INI file with this structure:
    [section]
    # comment
    ; comment
    VARIABLE = VALUE

    Ordered dict like:
    ==> [(section_name,
            OrderedDict([(VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         ...
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment))])),
         (section_name,
            OrderedDict([(VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         ...
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment))])),
         ...]

    :param dict_config: :OrderedDict: Content of INI file after 'readlines()'
    :param dest_path: :str: optional path for INI file write.
    :return: :str: raw INI text

    """
    lines = ['# -*- coding: utf-8 -*-']
    for section, entries in dict_config.items():
        lines.append('[{}]'.format(section.upper()))
        for variable_name, (value, var_type, comment) in entries.items():
            if comment:
                [lines.append('# {}'.format(l_wrap)) for l_wrap in textwrap.wrap(comment, 80)]
            lines.append('{} = {}'.format(variable_name, value))
        lines.append('')  # Separador entre secciones
    ini_text = '\n'.join(lines)
    if dest_path is not None:
        _makedirs(dest_path)
        with open(dest_path, 'w') as f:
            f.write(ini_text)
    return ini_text


def check_resource_files(dest_path, origin_path=None):
    """
    Check needed files and directories in DATA_PATH. Init if needed (1º exec).

    :param dest_path:
    :param origin_path:
    :return bool (dest_path exists previously)
    """
    if not os.path.exists(dest_path):
        _makedirs(dest_path)
        if origin_path is None:
            log('-> Made paths to "{}"'.format(dest_path), 'info', True, True)
        else:
            origin_path = os.path.abspath(origin_path)
            if os.path.isfile(origin_path):
                log('-> Copying resource file from "{}" to "{}"'.format(origin_path, dest_path), 'info', True, True)
                shutil.copy(origin_path, dest_path)
            else:
                log('-> Replicating tree from "{}" to "{}"'.format(origin_path, dest_path), 'info', True, True)
                shutil.copytree(origin_path, dest_path)
        log('** check_resource_files OK', 'debug', True, True)
        return False
    elif oct(os.stat(dest_path).st_mode)[-3:] != '777':
        os.chmod(dest_path, 0o777)
    return True


def _get_config():
    """
    Loads or generates ini file for ENERPI (& ENERPIweb) configuration.

    1) Looks for variable path DATA_PATH in file .../enerpi/config/.enerpi_data_path
    2) Tries to load 'config_enerpi.ini' from DATA_PATH, as user custom config.
    3) If not present, generates it copying the default configuration.

    :return: :configparser: loaded object
    """

    # Load DATA_PATH:
    dir_config = os.path.join(BASE_PATH, 'config')
    path_default_datapath = os.path.join(dir_config, '.enerpi_data_path')
    try:
        with open(path_default_datapath, 'r', encoding=ENCODING) as f:
            raw = f.read()
        data_path = raw.split('\n')[0]
        if data_path != os.path.expanduser(data_path):  # Hay '~', se expande el usuario y se graba abspath
            data_path = os.path.expanduser(data_path)
            log('''Sobreescritura del archivo "{}",
almacenando la ruta absoluta a la instalación de ENERPI
    -> DATA_PATH := {}
** Para mover la instalación de ENERPI a otro lugar, edite directamente este fichero
   (y mueva manualmente la carpeta DATA_PATH)'''.format(path_default_datapath, data_path), 'info', True, True)
            with open(path_default_datapath, 'w', encoding=ENCODING) as f:
                f.write(data_path)
    except Exception as e:
        log('ENERPI LOAD CONFIG ERROR at "{}" --> {} [{}]'
            .format(path_default_datapath, e, e.__class__), 'error', True, True)
        data_path = os.path.expanduser('~/ENERPIDATA')

    # Checks paths & re-gen if not existent
    check_resource_files(data_path)
    path_file_config = os.path.join(data_path, CONFIG_FILENAME)
    if not os.path.exists(path_file_config):
        check_resource_files(path_file_config)
        log('** Instalando fichero de configuración en: "{}"'.format(path_file_config), 'info', True, True)
        shutil.copy(os.path.join(dir_config, 'default_config_enerpi.ini'), path_file_config)

    # Config parser
    configp = configparser.RawConfigParser()
    try:
        configp.read(path_file_config, encoding=ENCODING)
    except Exception as e:
        log('Error loading configuration INI file in "{}". Exception {} [{}]. Using defaults...'
            .format(path_file_config, e, e.__class__), 'error', True, True)
        configp.read(os.path.join(dir_config, 'default_config_enerpi.ini'), encoding=ENCODING)
    return data_path, configp


# Loads configuration
DATA_PATH, CONFIG = _get_config()
FILE_LOGGING = os.path.join(DATA_PATH, CONFIG.get('ENERPI_DATA', 'FILE_LOGGING', fallback='enerpi.log'))
LOGGING_LEVEL = CONFIG.get('ENERPI_DATA', 'LOGGING_LEVEL', fallback='DEBUG')

# Set Locale
CUSTOM_LOCALE = CONFIG.get('ENERPI_SAMPLER', 'LOCALE', fallback='{}.{}'.format(*locale.getlocale()))
locale.setlocale(locale.LC_ALL, CUSTOM_LOCALE)

# ANALOG SENSORS WITH MCP3008 (Rasp.io Analog Zero)
SENSORS = EnerpiSamplerConf(CONFIG)



# -*- coding: utf-8 -*-
from collections import OrderedDict
from flask import request, redirect, url_for, render_template, send_file, abort
import json
import os
from enerpi.base import (DATA_PATH, CONFIG_FILENAME, FILE_LOGGING, CONFIG,
                         get_lines_file, log, make_ini_file, config_dict_for_web_edit, config_changes)
from enerpi.api import enerpi_data_catalog, get_encryption_key, get_codec
from enerpiweb import app, SERVER_FILE_LOGGING, STATIC_PATH


ENERPI_FILE_LOGGING = FILE_LOGGING
RSC_GEN_FILE_LOGGING = os.path.join(STATIC_PATH, 'enerpiweb_rscgen.log')


# Interesting files / logs to show:
def _get_filepath_from_file_id(file_id):
    if 'flask' == file_id:
        is_logfile, filename = True, SERVER_FILE_LOGGING
    elif 'rsc' == file_id:
        is_logfile, filename = True, RSC_GEN_FILE_LOGGING
    elif 'nginx_err' == file_id:
        is_logfile, filename = True, '/var/log/nginx/error.log'
    elif 'nginx' == file_id:
        is_logfile, filename = True, '/var/log/nginx/access.log'
    elif 'enerpi' == file_id:
        is_logfile, filename = True, ENERPI_FILE_LOGGING
    elif 'uwsgi' == file_id:
        is_logfile, filename = True, '/var/log/uwsgi/enerpiweb.log'
    elif 'config' == file_id:
        is_logfile, filename = False, os.path.join(DATA_PATH, CONFIG_FILENAME)
    elif 'encryption_key' == file_id:
        is_logfile, filename = False, os.path.join(DATA_PATH,
                                                   CONFIG.get('BROADCAST', 'KEY_FILE', fallback='.secret_key'))
    else:  # Fichero derivado del catálogo
        cat = enerpi_data_catalog(check_integrity=False)
        if 'raw_store' == file_id:
            is_logfile, filename = False, os.path.join(cat.base_path, cat.raw_store)
        elif 'catalog' == file_id:
            is_logfile, filename = False, os.path.join(cat.base_path, cat.catalog_file)
        else:
            log('FILE_ID No reconocido: {}'.format(file_id), 'error', False)
            is_logfile, filename = False, SERVER_FILE_LOGGING
            return False, is_logfile, filename
    return True, is_logfile, filename


#############################
# ROUTES for file handling
#############################
@app.route('/api/hdfstores/<relpath_store>', methods=['GET'])
def download_hdfstore_file(relpath_store=None):
    """
    Devuelve el fichero HDFStore del catálogo de ENERPI pasado como ruta relativa (nombre de fichero .h5).
    * File Download *
    :param relpath_store:
    """
    cat = enerpi_data_catalog(check_integrity=False)
    path_file = cat.get_path_hdf_store_binaries(relpath_store)
    if 'as_attachment' in request.args:
        return send_file(path_file, as_attachment=True, attachment_filename=os.path.basename(path_file))
    return send_file(path_file, as_attachment=False)


@app.route('/api/filedownload/<file_id>', methods=['GET'])
def download_file(file_id):
    """
    * File Download *
    :param file_id:
    """
    ok, _, filename = _get_filepath_from_file_id(file_id)
    if ok:
        if os.path.exists(filename):
            if 'as_attachment' in request.args:
                return send_file(filename, as_attachment=True, attachment_filename=os.path.basename(filename))
            return send_file(filename, as_attachment=False)
        else:
            msg = json.dumps({'alert_type': 'danger',
                              'texto_alerta': 'El archivo "{}" ({}) no existe!'.format(filename, file_id)})
            return redirect(url_for('control', alerta=msg))
    return abort(404)


@app.route('/api/showfile')
@app.route('/api/showfile/<file>')
def showfile(file='flask'):
    """
    Página de vista de fichero de texto, con orden ascendente / descendente y/o nº de últimas líneas ('tail' de archivo)
    :param file: file_id to show
    """

    ok, is_logfile, filename = _get_filepath_from_file_id(file)
    if ok:
        delete = request.args.get('delete', '')
        reverse = request.args.get('reverse', False)
        tail_lines = request.args.get('tail', None)
        alerta = request.args.get('alerta', '')
        if alerta:
            alerta = json.loads(alerta)
        if not alerta and delete and is_logfile:
            with open(filename, 'w') as f:
                f.close()
            cad_delete = 'LOGFILE {} DELETED'.format(filename.upper())
            log(cad_delete, 'warn', False)
            return redirect(url_for('showfile', file=file,
                                    alerta=json.dumps({'alert_type': 'warning', 'texto_alerta': cad_delete})))
        data = get_lines_file(filename, tail=tail_lines, reverse=reverse)
        return render_template('text_file.html', titulo='LOG File:' if is_logfile else 'CONFIG File:', file_id=file,
                               subtitulo='<strong>{}</strong>'.format(filename), is_logfile=is_logfile,
                               file_content=data, filename=os.path.basename(filename), alerta=alerta)
    return abort(404)


@app.route('/api/editconfig/', methods=['GET'])
@app.route('/api/editconfig/<file>', methods=['GET', 'POST'])
def editfile(file='config'):
    """
    Config editor, for INI file & encripting key

    :param file: file_id to edit
    """
    # TODO readonly/disabled attr in protected fields of INI file

    ok, is_logfile, filename = _get_filepath_from_file_id(file)
    if ok:
        if is_logfile:
            return redirect(url_for('index'), code=404)

        alerta = request.args.get('alerta', '')
        if alerta:
            alerta = json.loads(alerta)
        without_comments = request.args.get('without_comments', 'False')
        without_comments = without_comments.lower() == 'true'

        if file == 'encryption_key':  # Key edition!! NOT SAFE AT ALL (Debugging purposes!)
            extra_links = [('<i class="fa fa-check-square-o" aria-hidden="true"></i> Edit Configuration',
                            url_for('editfile', file='config'))]
            show_switch_comments = False
            without_comments = True
            variable_name = 'KEY_FILE'
            key = get_encryption_key(filename)
            config_data = OrderedDict(
                [('Encryption KEY', OrderedDict([(variable_name, (key.decode(), 'text', None))]))])
            lines_config = [key.decode()]
            if request.method == 'POST':
                new_key = request.form[variable_name]
                # if new_key.encode() != key:
                if new_key != key.decode():
                    try:
                        _ = get_codec(new_key.encode())
                        if len(new_key) > 10:
                            log('The new KEY ("{}")\nwill be saved in "{}"\n'.format(new_key, filename), 'warning',
                                True)
                            with open(filename, 'wb') as f:
                                f.write(new_key.encode())
                            str_cambios = ('Configuration changes in encryption key:<br>{}<br> New config SAVED!'
                                           .format('- <strong>"{}"</strong> (before={}) -> <strong>"{}"</strong>'
                                                   .format(variable_name, key.decode(), new_key)))
                            alerta = {'alert_type': 'warning', 'texto_alerta': str_cambios}
                            lines_config = [new_key]
                            config_data = OrderedDict([('Encryption KEY',
                                                        OrderedDict([(variable_name, (new_key, 'text', None))]))])
                    except AssertionError as e:
                        alerta = {'alert_type': 'danger', 'texto_alerta': '{}. New KEY was: {}'.format(e, new_key)}
        else:  # INI CONFIG FILE EDITION:
            extra_links = [('<i class="fa fa-user-secret" aria-hidden="true"></i> Edit SECRET Key',
                            url_for('editfile', file='encryption_key'))]
            show_switch_comments = True
            lines_config = get_lines_file(filename)
            config_data = config_dict_for_web_edit(lines_config)
            if request.method == 'POST':
                hay_cambio, vars_updated, config_data_updated = config_changes(request.form, config_data)
                if hay_cambio:
                    str_cambios = ('Configuration changes in:<br>{}<br> New config SAVED!'
                                   .format('<br>'.join(['- <strong>"{}"</strong> (before={}) -> <strong>"{}"</strong>'
                                                       .format(name, ant, new) for name, ant, new in vars_updated])))
                    alerta = {'alert_type': 'warning', 'texto_alerta': str_cambios}
                    lines_config = make_ini_file(config_data_updated, dest_path=filename).splitlines()
                    config_data = config_data_updated
        return render_template('edit_text_file.html', titulo='CONFIGURATION EDITOR', file_id=file,
                               show_switch_comments=show_switch_comments, with_comments=not without_comments,
                               abspath=filename, dict_config_content=config_data, file_lines=lines_config,
                               filename=os.path.basename(filename), alerta=alerta, extra_links=extra_links)
    return abort(404)


@app.route('/api/uploadfile/<file>', methods=['POST'])
def uploadfile(file):
    """
    POST method for VIP files upload & replacement

    :param file: uploaded file_id
    """
    ok, is_logfile, filename = _get_filepath_from_file_id(file)
    if ok:
        if is_logfile:
            print('uploading logfile!!!')
            return redirect(url_for('index'), code=404)
        # filename = secure_filename(filename)
        if len(filename) > 0:
            f = request.files['file']
            if f.mimetype == 'application/octet-stream':
                log('UPLOADED FILE id={}, filename="{}" (dest_path="{}")'.format(file, f.filename, filename), 'warn')
                f.save(filename)
                msg = json.dumps({'alert_type': 'success', 'texto_alerta': ('Uploaded file "{}" as replacement in "{}"!'
                                                                            .format(f.filename, filename))})
                return redirect(url_for('editfile', file=file, alerta=msg))
    return abort(500)


# TODO Implementar generación de PDF reports
# from flask_weasyprint import HTML, render_pdf
#
#
# @app.route('/api/filetopdf/<file>')
# def filetopdf(file='flask'):
#     """
#     Página de vista de fichero de texto, con orden ascendente / descendente y/o nº de últimas líneas
#     ('tail' de archivo)
#     :param file: file_id to show
#     """
#     return render_pdf(url_for('showfile', file=file))



# -*- coding: utf-8 -*-
import argparse
import logging
import os
from time import time, sleep
# do this before importing pylab or pyplot
import matplotlib
matplotlib.use('Agg')
# noinspection PyPep8
from enerpi.base import DATA_PATH, CONFIG, check_resource_files, set_logging_conf
# noinspection PyPep8
from enerpi.database import init_catalog
# noinspection PyPep8
from enerpiplot.enerplot import gen_svg_tiles


basedir_enerweb = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', 'enerpiweb')
STATIC_PATH = os.path.join(DATA_PATH, CONFIG.get('ENERPI_WEBSERVER', 'STATIC_PATH'))
SERVER_FILE_LOGGING_RSCGEN = os.path.join(STATIC_PATH, 'enerpiweb_rscgen.log')
IMG_TILES_BASEPATH = os.path.join(STATIC_PATH, 'img', 'generated')
check_resource_files(STATIC_PATH, os.path.join(basedir_enerweb, 'static'))
check_resource_files(IMG_TILES_BASEPATH)
TILES_GENERATION_INTERVAL = 180
LOGGING_LEVEL_SERVER = 'DEBUG'

# Establecemos logging
set_logging_conf(SERVER_FILE_LOGGING_RSCGEN, LOGGING_LEVEL_SERVER, with_initial_log=False)
# logging.debug('(MULE_RSCGEN)->LOG Estableciendo LOGFILE en {}'.format(SERVER_FILE_LOGGING_RSCGEN))


###############################
# RESOURCE GENERATOR
###############################
def _rsc_generator(catalog):
    tic, ok = time(), False
    # logging.debug('**RESOURCE_GENERATOR desde MULE con PID={}!!!'.format(os.getpid()))
    # try:
    ok = gen_svg_tiles(IMG_TILES_BASEPATH, catalog, last_hours=(72, 48, 24))
    toc = time()
    logging.debug('(MULE) TILES generation ok? {}. TOOK {:.3f} s'.format(ok, toc - tic))
    # except Exception as e:
    #     toc = time()
    #     logging.error('Error {} [{}] en gen_svg_tiles. TOOK {:.3f} s'.format(e, e.__class__, toc - tic))
    return ok, toc - tic


def _loop_rsc_generator(catalog, sleep_time):
    logging.debug('**RESOURCE_GENERATOR LOOP desde MULE con PID={}!!!'.format(os.getpid()))
    counter = 0
    while True:
        gen_ok, took = _rsc_generator(catalog)
        counter += 1
        sleep(max(sleep_time - took, 20))
        # uwsgi.async_sleep(max(sleep_time - took, 20))


def main():
    """
    CLI Manager for generate svg plots as resources for the web app
    """
    p = argparse.ArgumentParser(description="EnerPI Web Resource generator")
    p.add_argument('-o', '--one', action='store_true', help='Run only one time.')
    p.add_argument('-t', '--timer', action='store', type=int, metavar='∆T', default=TILES_GENERATION_INTERVAL,
                   help='Set periodic timer, in seconds. Default={}.'.format(TILES_GENERATION_INTERVAL))
    args = p.parse_args()
    # logging.debug('(MULE)-> main con args={}'.format(args))

    # Cargamos catálogo para consulta:
    cat = init_catalog(check_integrity=False)

    if args.one:
        _rsc_generator(cat)
    else:
        logging.debug('(MULE)-> Init loop with timer={} seconds'.format(args.timer))
        # print('(MULE)-> Init loop with timer={} seconds'.format(args.timer))
        _loop_rsc_generator(cat, args.timer)
    # logging.debug('(MULE) Saliendo')
    # p = Process(target=_rsc_generator, args=(cat,))
    # p.start()
    # p.join()
    # _rsc_generator(cat)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
** CRON Scheduler **
"""
from crontab import CronTab
import datetime as dt


def info_crontable():
    """
    Print current user CRON TABLE

    """
    ahora = dt.datetime.now()
    cron = CronTab(user=True)
    print('-*-' * 30)
    for job in cron:
        # Validity Check:
        assert(job.is_valid())
        schedule = job.schedule(date_from=ahora)
        datetime_n = schedule.get_next()
        print('** JOB: {}'.format(job))
        print('nº execs this year: {}'.format(job.frequency_per_year(year=ahora.year)))
        print('NEXT: {:%H:%M %d-%m-%Y}'.format(datetime_n))
        print('PREV: {:%H:%M %d-%m-%Y}'.format(schedule.get_prev()))
        print('*' * 90)
    print('-*-' * 30)


def set_command_on_reboot(command, comment=None, verbose=True):
    """
    Sets CRON task in current user crontab (like editing it with 'crontab -e') at @reboot
    :param command: :str: CRON command
    :param comment: :str: comment about the command
    :param verbose: :bool: shows info about the CRON job
    """
    cron = CronTab(user=True)
    for job in cron.find_command(command):
        cron.remove(job)
    if comment is None:
        new_job = cron.new(command=command)
    else:
        new_job = cron.new(command=command, comment=comment)
    new_job.every_reboot()
    if verbose:
        print('->CRON JOB: {}'.format(new_job))
        # print('job.is_valid = {}'.format(job.is_valid()))
    # Validity Check:
    assert(new_job.is_valid())
    # Write CronTab back to system or filename:
    cron.write()


def set_command_periodic(command, comment=None, hour=None, minute=None, frecuency_days=None, verbose=True):
    """
    Sets CRON task in current user crontab (like editing it with 'crontab -e') at some frequency
    :param command: :str: CRON command
    :param comment: :str: comment about the command
    :param hour: :int or None: every X hours
    :param minute: :int or None: every X minutes
    :param frecuency_days: :int or None: every X days
    :param verbose: :bool: shows info about the CRON job
    """
    cron = CronTab(user=True)
    for job in cron.find_command(command):
        cron.remove(job)
    if comment is None:
        new_job = cron.new(command=command)
    else:
        new_job = cron.new(command=command, comment=comment)
    if hour is not None:
        new_job.hour.every(hour)
    if minute is not None:
        new_job.minute.every(minute)
    if frecuency_days is not None:
        new_job.day.every(frecuency_days)
    assert(new_job.is_valid())

    schedule = new_job.schedule(date_from=dt.datetime.now())
    datetime_n = schedule.get_next()
    if verbose:
        print('->CRON JOB: {}\n  * NEXT PROGRAMMED EXEC: {:%H:%M %d-%m-%Y}'.format(new_job, datetime_n))
    # Write CronTab back to system or filename:
    cron.write()
    return datetime_n


def clear_cron_commands(commands, verbose=True):
    """
    Deletes a list of CRON tasks (as list of commands)
    :param commands: list of commands to delete
    :param verbose: :bool: shows info about the operation
    :return:
    """
    cron = CronTab(user=True)
    for command in commands:
        for job in cron.find_command(command):
            if verbose:
                print(' --> DELETING CRON-JOB: "{}"'.format(job))
            cron.remove(job)
    cron.write()
    return True


if __name__ == '__main__':
    info_crontable()



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
import os
import sys
from enerpi.base import log, FILE_LOGGING, LOGGING_LEVEL, set_logging_conf
from enerpi.enerpimeter import enerpi_daemon_logger
from enerpi.daemon import Daemon


class EnerPiDaemon(Daemon):
    """
    Clase Daemon para la ejecución de enerpi_main_logger en background
    """
    def run(self):
        """
        Runs the logger
        """
        enerpi_daemon_logger()


def enerpi_daemon(test_mode=False):
    """
    Main logic para el demonio de ejecución de enerpi_main_logger

    Ej de uso:
    sudo -u www-data /home/pi/PYTHON/py35/bin/enerpi-daemon start|stop|status|restart
    enerpi-daemon start|stop|status|restart
    Opciones:
        - start
        - stop
        - status
        - restart
    """
    set_logging_conf(FILE_LOGGING, LOGGING_LEVEL, with_initial_log=False)
    # TODO Eliminar stdout y stderr! (o diferenciar su uso según LOGGING_LEVEL)
    daemon = EnerPiDaemon('/tmp/enerpilogger.pid',
                          test_mode=test_mode,
                          stdout=os.path.expanduser('/tmp/enerpi_out.txt'),
                          stderr=os.path.expanduser('/tmp/enerpi_err.txt'))
    if len(sys.argv) == 2:
        if 'start' == sys.argv[1]:
            log('ENERPI Logger daemon started', 'info', False)
            try:
                daemon.start()
            except PermissionError as e:
                sys.stderr.write("PERMISSIONERROR: pidfile can't be registered ({}). Need sudo powers?".format(e))
        elif 'stop' == sys.argv[1]:
            log('ENERPI Logger daemon stopped', 'warn', False)
            daemon.stop()
        elif 'status' == sys.argv[1]:
            log('ENERPI Logger daemon status?', 'debug', False)
            daemon.status()
        elif 'restart' == sys.argv[1]:
            log('ENERPI Logger daemon restarting', 'info', False)
            daemon.restart()
        else:
            log("Unknown command", 'warn', True, False)
            if not test_mode:
                sys.exit(2)
        if not test_mode:
            sys.exit(0)
    else:
        print("usage: %s start|stop|restart|status".format(sys.argv[0]))
        if not test_mode:
            sys.exit(2)


if __name__ == "__main__":
    enerpi_daemon()



# -*- coding: utf-8 -*-
import datetime as dt
from io import BytesIO
import locale
import logging
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
import matplotlib.dates as mpd
import matplotlib.patches as mp
import numpy as np
import os
import re
from enerpi.base import CONFIG, DATA_PATH, CUSTOM_LOCALE, timeit


IMG_BASEPATH = os.path.join(DATA_PATH, CONFIG.get('ENERPI_DATA', 'IMG_BASEPATH'))
DEFAULT_IMG_MASK = CONFIG.get('ENERPI_DATA', 'DEFAULT_IMG_MASK')


def _gen_tableau20():
    # # These are the "Tableau 20" colors as RGB.
    tableau = [(31, 119, 180), (174, 199, 232), (255, 127, 14), (255, 187, 120),
               (44, 160, 44), (152, 223, 138), (214, 39, 40), (255, 152, 150),
               (148, 103, 189), (197, 176, 213), (140, 86, 75), (196, 156, 148),
               (227, 119, 194), (247, 182, 210), (127, 127, 127), (199, 199, 199),
               (188, 189, 34), (219, 219, 141), (23, 190, 207), (158, 218, 229)]
    # Scale the RGB values to the [0, 1] range, which is the format matplotlib accepts.
    for i in range(len(tableau)):
        r, g, b = tableau[i]
        tableau[i] = (r / 255., g / 255., b / 255.)
    return tableau


# semaforo_4 = [sns.palettes.crayons[k] for k in ['Green', 'Sea Green', 'Mango Tango', 'Razzmatazz']]
# These are the "Tableau 20" colors as RGB.
tableau20 = _gen_tableau20()
locale.setlocale(locale.LC_ALL, CUSTOM_LOCALE)
# sns.set_style('whitegrid')

REGEXPR_SVG_HEIGHT = re.compile(r'<svg height="\d{1,4}pt"')
REGEXPR_SVG_WIDTH = re.compile(r' width="(\d{1,4}pt")')

GRIDSPEC_FULL = {'left': 0, 'right': 1, 'bottom': 0, 'top': 1, 'hspace': 0}
# GRIDSPEC_NORMAL = {'left': 0.075, 'right': .925, 'bottom': 0.11, 'top': 0.91, 'hspace': 0}
FONTSIZE = 10
FONTSIZE_TILE = 12
TICK_PARAMS_TILE = dict(direction='in', pad=-15, length=3, width=.5)
font = {'family': 'sans-serif',
        'size': FONTSIZE}  # 'weight' : 'light',
matplotlib.rc('font', **font)


def ch_color(x, ch=1., alpha=None):
    """
    Modify color applying a multiplier in every channel, or setting an alpha value

    :param x:  :tuple: 3/4 channel normalized color (0->1. * 3/4 ch)
    :param ch: :float: change coefficient
    :param alpha: :float: alpha value
    :return: :tuple: modified color

    """
    new_c = [max(0, min(1., ch * c)) for c in x]
    if alpha is not None:
        if len(x) == 4:
            new_c[3] = alpha
            return tuple(new_c)
        else:
            return tuple(new_c + [alpha])
    return tuple(new_c)


def _round_time(ts=None, delta=dt.timedelta(minutes=1)):
    round_to = delta.total_seconds()
    if ts is None:
        ts = dt.datetime.now()
    seconds = (ts - ts.min).seconds
    rounding = (seconds + round_to / 2) // round_to * round_to
    return ts + dt.timedelta(0, rounding - seconds, -ts.microsecond)


def _fix_time_series_bar_mix(ax_ts, ax_bar, rango_ts, index_ts_bar, bar_width=.8):
    barras = [c for c in ax_bar.get_children() if type(c) is mp.Rectangle]
    delta_ts = rango_ts[1] - rango_ts[0]
    rango = ax_ts.get_xlim()
    ax_ts.set_xlim(rango)
    delta = rango[1] - rango[0]
    pos_start_barras = [((ts_b - rango_ts[0]) / delta_ts) * delta + rango[0] for ts_b in index_ts_bar]
    ax_ts.set_xticks(pos_start_barras)
    ax_bar.set_xticks(pos_start_barras)
    max_width = pos_start_barras[1] - pos_start_barras[0]
    new_width = bar_width * max_width
    if bar_width != 1.:
        pos_start_barras = [x + (1 - bar_width) * max_width / 2 for x in pos_start_barras]
    for b, xs in zip(barras[:-1], pos_start_barras):
        b.set_x(xs)
        b.set_width(new_width)
    ax_bar.set_xlim(rango_ts)
    ax_ts.set_xlim(rango_ts)


def _format_timeseries_axis(ax_bar, ax_ts, rango_ts,
                            fmt='%-H:%M', ini=0, fin=None, axis_label='',
                            fmt_mayor=" %-d %h'%y, %-Hh", mayor_divisor=12,
                            fmt_mayor_day="%A %-d %h'%y",
                            delta_ticks=dt.timedelta(minutes=0)):

    def _gen_mayor(x, fmt_label, fmt_label_d):
        label = ' ' + x.strftime(fmt_label_d).capitalize() if x.hour == 0 else x.strftime(fmt_label)
        return mpd.date2num(x), label

    delta_day = dt.timedelta(days=1)
    if fin is None:
        xticks = ax_bar.get_xticks()[ini:]
    else:
        xticks = ax_bar.get_xticks()[ini:fin]
    new_xticks_dt = [_round_time(dt.datetime.fromordinal(int(x)) + delta_day * (x % 1), dt.timedelta(hours=1))
                     for x in xticks]
    new_xticks, new_ts_labels = list(zip(*[(mpd.date2num(x + delta_ticks), x.strftime(fmt)) for x in new_xticks_dt]))
    color_mayor = ch_color(tableau20[14], alpha=.7)
    color_minor = ch_color(tableau20[14], 1.1, alpha=.7)
    ax_bar.set_xticks(new_xticks)
    ax_bar.set_xticklabels(new_ts_labels, rotation=0, ha='center', fontweight='bold', fontsize='medium')
    ax_bar.tick_params(axis='x', direction='out', length=2, width=2, pad=5, bottom='on', top='off',
                       color=color_minor, labelcolor=color_minor)
    ax_bar.set_xlabel(axis_label)
    ax_bar.grid(False, axis='x')
    xticks_mayor, ts_labels_mayor = list(zip(*[_gen_mayor(x, fmt_mayor, fmt_mayor_day)
                                               for x in filter(lambda x: x.hour % mayor_divisor == 0, new_xticks_dt)]))
    ax_ts.set_xlabel('')
    ax_ts.xaxis.set_ticks_position('top')
    ax_ts.yaxis.set_ticks_position('right')
    ax_ts.yaxis.set_label_position('right')
    ax_ts.set_xticks(xticks_mayor)
    ax_ts.set_xticklabels(ts_labels_mayor, rotation=0, ha='left', fontweight='bold', fontsize='large')
    ax_ts.tick_params(axis='x', direction='out', length=20, width=2, pad=5,
                      color=color_mayor, labelcolor=color_mayor)
    ax_ts.grid(True, axis='x', color=color_mayor, linestyle='-', linewidth=2, alpha=.7)
    ax_bar.set_xlim(rango_ts)
    ax_ts.set_xlim(rango_ts)


def _gen_image_path(data, filename):
    if type(filename) is str:
        if filename.lower() in ['png', 'svg', 'pdf', 'jpeg']:
            img_name = DEFAULT_IMG_MASK[:-3] + filename
        else:
            img_name = filename
    else:
        img_name = DEFAULT_IMG_MASK
    if len(os.path.splitext(img_name)[1]) == 0:
        img_name += '.png'
    head, tail = os.path.split(img_name)
    if not head:
        img_name = os.path.join(IMG_BASEPATH, img_name)
    masks = img_name.count('{:')
    if masks == 2:
        return img_name.format(data.index[0], data.index[-1])
    elif masks == 1:
        return img_name.format(data.index[0])
    else:
        return img_name


@timeit('plot_power_consumption_hourly')
def plot_power_consumption_hourly(potencia, consumo, ldr=None, rs_potencia=None, rm_potencia=None, savefig=None):
    f, ax_bar = plt.subplots(figsize=(16, 9))
    color_potencia = ch_color(tableau20[4], .85, alpha=.9)
    color_consumo = ch_color(tableau20[8], alpha=.9)
    color_ldr = ch_color(tableau20[2], alpha=.6)

    params_bar = dict(ax=ax_bar, kind='bar', color=ch_color(tableau20[9], alpha=.6),
                      lw=1.5, edgecolor=color_consumo)
    ax_bar = consumo.plot(**params_bar)

    ax_pos = ax_bar.get_position()
    ax_ts = f.add_axes(ax_pos, frameon=False, axis_bgcolor=None)

    if rm_potencia is not None:
        potencia = potencia.rolling(rm_potencia).mean().fillna(0)
        if ldr is not None:
            ldr = ldr.rolling(rm_potencia).mean().fillna(0)
    elif rs_potencia is not None:
        potencia = potencia.resample(rs_potencia, label='left').mean().fillna(0)
        if ldr is not None:
            ldr = ldr.resample(rs_potencia, label='left').mean().fillna(0)
    rango_ts = potencia.index[0], potencia.index[-1]
    ax_ts.plot(potencia, lw=1, color=color_potencia)
    ax_ts.fill_between(potencia.index, potencia.values, 0, lw=0, facecolor=ch_color(tableau20[5], alpha=.4),
                       hatch='\\', edgecolor=color_potencia, zorder=1)

    ratio = 2000
    grid_w = 1000
    ylim_c = np.ceil(ax_bar.get_ylim()[1] / .5) * .5
    ylim_p = np.ceil(ax_ts.get_ylim()[1] / grid_w) * grid_w
    ylim = max(ylim_c * ratio, ylim_p)

    if ldr is not None:
        ax_ts.plot(ldr * ylim / 1000, lw=1, color=color_ldr, zorder=0)
        ax_ts.fill_between(ldr.index, ldr.values * ylim / 1000, 0, lw=0,
                           facecolor=ch_color(tableau20[3], alpha=.1), zorder=0)

    # Remove plot frame
    ax_ts.set_frame_on(False)

    # Formating xticks
    _fix_time_series_bar_mix(ax_ts, ax_bar, rango_ts, consumo.index, bar_width=.8)
    _format_timeseries_axis(ax_bar, ax_ts, rango_ts)  # , ini=1, fin=-1)

    # Formating Y axes:
    ax_bar.set_ylabel('Consumo eléctrico', ha='center', fontweight='bold', fontsize='x-large', color=color_consumo)
    ax_ts.set_ylabel('Potencia eléctrica', ha='center', fontweight='bold', fontsize='x-large', color=color_potencia,
                     rotation=270, labelpad=20)
    ax_bar.spines['left'].set_color(color_consumo)
    ax_bar.spines['left'].set_linewidth(2)
    ax_bar.spines['right'].set_color(color_potencia)
    ax_bar.spines['right'].set_linewidth(2)
    ax_bar.spines['bottom'].set_linewidth(0)
    ax_ts.hlines(0, *rango_ts, colors=ch_color(tableau20[14], 1.3), lw=5)
    yticks_p = np.linspace(grid_w, ylim, ylim / grid_w)
    yticks_c = np.linspace(grid_w / ratio, ylim / ratio, ylim / grid_w)
    yticklabels_p = ['{:.2g}'.format(x / 1000.) + ' kW' for x in yticks_p]
    yticklabels_c = ['{:.2g}'.format(x) + ' kWh' for x in yticks_c]
    ax_bar.set_yticks(yticks_c)
    ax_bar.set_yticklabels(yticklabels_c, ha='right', fontweight='bold', fontsize='large')
    ax_bar.tick_params(axis='y', direction='out', length=5, width=2, pad=10,
                       color=color_consumo, labelcolor=color_consumo)
    ax_ts.set_yticks(yticks_p)
    ax_ts.set_yticklabels(yticklabels_p, ha='left', fontweight='bold', fontsize='large')
    ax_ts.tick_params(axis='y', direction='out', length=5, width=2, pad=10,
                      color=color_potencia, labelcolor=color_potencia)

    ax_bar.set_position(ax_pos)
    if savefig is not None:
        img_name = _gen_image_path(potencia, savefig)
        f.savefig(img_name, dpi=300, transparent=True, bbox_inches='tight', pad_inches=0.1, frameon=False)
        return img_name
    else:
        return f, [ax_bar, ax_ts]


# @timeit('write_fig_to_svg', verbose=True)
def write_fig_to_svg(fig, name_img, preserve_ratio=False):
    """
    Write matplotlib figure to disk in SVG format.

    :param fig: :matplotlib.figure: figure
    :param name_img: :str: desired image path
    :param preserve_ratio: :bool: preserve size ratio on the SVG file (default: False)
    :return: :book: operation ok

    """
    canvas = FigureCanvas(fig)
    output = BytesIO()
    imgformat = 'svg'
    canvas.print_figure(output, format=imgformat, transparent=True)
    svg_out = output.getvalue()
    if not preserve_ratio:
        svg_out = REGEXPR_SVG_WIDTH.sub(' width="100%" preserveAspectRatio="none"',
                                        REGEXPR_SVG_HEIGHT.sub('<svg height="100%"', svg_out.decode())).encode()
    try:
        with open(name_img, 'wb') as f:
            f.write(svg_out)
    except Exception as e:
        logging.error('HA OCURRIDO UN ERROR GRABANDO SVG A DISCO: {}'.format(e))
        print('HA OCURRIDO UN ERROR GRABANDO SVG A DISCO: {}'.format(e))
        return False
    return True


def _tile_figsize(fraction=1.):
    dpi = 72
    # height = 200
    # width = 1.875 * height
    height = 200
    width = 4.5 * height * fraction
    return round(width / dpi, 2), round(height / dpi, 2)


def _prep_axis_tile(color):
    fig, ax = plt.subplots(figsize=_tile_figsize(), dpi=72, gridspec_kw=GRIDSPEC_FULL, facecolor='none')
    fig.patch.set_alpha(0)
    ax.patch.set_alpha(0)
    ax.tick_params(direction='in', pad=-15, length=3, width=.5)
    ax.tick_params(axis='y', length=0, width=0, labelsize=FONTSIZE_TILE)
    ax.tick_params(axis='x', which='both', top='off', labelbottom='off')
    ax.xaxis.grid(True, color=color, linestyle=':', linewidth=1.5, alpha=.6)
    ax.yaxis.grid(True, color=color, linestyle=':', linewidth=1, alpha=.5)
    return fig, ax


def _adjust_tile_limits(name, ylim, date_ini, date_fin, ax):
    ax.set_ylim(ylim)
    ax.set_xlim(left=date_ini, right=date_fin)
    yticks = list(ax.get_yticks())[1:-1]
    yticks_l = [v for v in yticks if (v - ylim[0] < (2 * (ylim[1] - ylim[0]) / 3)) and (v > ylim[0])]
    ax.set_yticks(yticks)
    if name == 'power':
        ax.set_yticklabels([str(round(float(y / 1000.), 1)) + 'kW' for y in yticks_l])
        ax.tick_params(pad=-45)
    elif name == 'ldr':
        ax.set_yticklabels([str(round(float(y / 10.))) + '%' for y in yticks_l])
        ax.tick_params(pad=-40)
    else:
        ax.tick_params(pad=-30)
        ax.set_yticklabels([str(round(y, 4)) for y in yticks_l])
    return ax


@timeit('plot_tile_last_24h')
def plot_tile_last_24h(data_s, barplot=False, ax=None, fig=None, color=(1, 1, 1), alpha=1, alpha_fill=.5):
    matplotlib.rcParams['axes.linewidth'] = 0
    if ax is None:
        fig, ax = _prep_axis_tile(color)
    else:
        ax.patch.set_alpha(0)
        ax.tick_params(direction='in', pad=-15, length=3, width=.5)
        ax.tick_params(axis='y', length=0, width=0, labelsize=FONTSIZE_TILE)
        ax.tick_params(axis='x', which='both', top='off', labelbottom='off')
        ax.xaxis.grid(True, color=color, linestyle=':', linewidth=1.5, alpha=.6)
        ax.yaxis.grid(True, color=color, linestyle=':', linewidth=1, alpha=.5)
    rango_ts = data_s.index[0], data_s.index[-1]
    date_ini, date_fin = [t.to_pydatetime() for t in rango_ts]
    if data_s is not None and not data_s.empty:
        lw = 1.5
        ax.grid(b=True, which='major')
        data_s = data_s.fillna(0)
        if barplot:
            div = .5
            ylim = (0, np.ceil((data_s.max() + div) // div) * div)
            ax.bar(data_s.index, data_s.values, width=1 / 28, edgecolor=color, color=list(color) + [.5], linewidth=lw)
            ax.xaxis.set_major_locator(mpd.HourLocator((0, 12)))
        else:
            if data_s.name == 'power':
                div = 500
                ylim = (0, np.ceil((data_s.max() + div / 5) / div) * div)
            else:  # ldr
                div = 100
                ylim = (0, np.ceil((data_s.max() + div) // div) * div)
            data_s = data_s
            ax.plot(data_s.index, data_s, color=color, linewidth=lw, alpha=alpha)
            ax.fill_between(data_s.index, data_s, color=color, alpha=alpha_fill)
            ax.xaxis.set_major_locator(mpd.HourLocator((0, 12)))
            ax.xaxis.set_minor_locator(mpd.HourLocator(interval=1))
    else:
        ylim = 0, 100
        ax.annotate('NO DATA!', xy=(.35, .3), xycoords='axes fraction',
                    va='center', ha='center', color=ch_color(color, .9), fontsize=25)

    _adjust_tile_limits(data_s.name, ylim, date_ini, date_fin, ax)
    return fig, ax


# TODO Parametrize generated tiles
def gen_svg_tiles(path_dest, catalog, last_hours=(72, 48, 24)):

    def _cut_axes_and_save_svgs(figure, axes, x_lim, delta_total, data_name):
        for lh in last_hours:
            file = os.path.join(path_dest, 'tile_{}_{}_last_{}h.svg'.format('enerpi_data', data_name, lh))
            axes.set_xlim((x_lim[0] + delta_total * (1 - lh / total_hours), x_lim[1]))
            figure.set_figwidth(_tile_figsize(lh / total_hours)[0])
            write_fig_to_svg(figure, name_img=file)

    total_hours = last_hours[0]
    last_data, last_data_c = catalog.get(last_hours=last_hours[0], with_summary=True)
    if (last_data is not None) and (len(last_data) > 2):
        ahora = dt.datetime.now().replace(second=0, microsecond=0)
        xlim = mpd.date2num(ahora - dt.timedelta(hours=last_hours[0])), mpd.date2num(ahora)
        delta = xlim[1] - xlim[0]

        fig, ax = plot_tile_last_24h(catalog.resample_data(last_data.power, rs_data='5min'), barplot=False)
        _cut_axes_and_save_svgs(fig, ax, xlim, delta, last_data.power.name)

        plt.cla()
        fig.set_figwidth(_tile_figsize()[0])
        fig, ax = plot_tile_last_24h(catalog.resample_data(last_data.ldr, rs_data='30s', use_median=True),
                                     barplot=False, ax=ax, fig=fig)
        _cut_axes_and_save_svgs(fig, ax, xlim, delta, last_data.ldr.name)

        if len(last_data_c) > 1:
            fig, ax = plot_tile_last_24h(last_data_c.kWh, barplot=True)
            _cut_axes_and_save_svgs(fig, ax, xlim, delta, last_data_c.kWh.name)

        if fig is not None:
            plt.close(fig)
        return True
    else:
        return False
